from __future__ import annotations

__all__ = ['ActivationFn', 'LazyConvFn', 'LazyNormFn', 'NameMixin', 'round8']

from typing import Protocol, TypeVar

from torch import nn

_T = TypeVar('_T')


def pair(t: _T | tuple[_T, ...]) -> tuple[_T, ...]:
    return t if isinstance(t, tuple) else (t, t)


class NameMixin:
    __constants__ = ['name']
    name: str

    def __repr__(self) -> str:
        return f'{type(self).__name__}({self.extra_repr()})'

    def extra_repr(self) -> str:
        return self.name


class LazyConvFn(Protocol):
    def __call__(self,
                 out_channels: int,
                 kernel_size: int,
                 stride: int = 1,
                 padding: int = 0,
                 dilation: int = 1,
                 groups: int = 1,
                 bias: bool = True) -> nn.modules.conv._ConvNd:
        ...


class ActivationFn(Protocol):
    def __call__(self, inplace: bool = ...) -> nn.Module:
        ...


class LazyNormFn(Protocol):
    def __call__(self) -> nn.Module:
        ...


def round8(v: float, divisor: int = 8) -> int:
    """Ensure that number rounded to nearest 8, and error is less than 10%

    This function is taken from the original tf repo.
    It ensures that all layers have a channel number that is divisible by 8
    It can be seen here:
    https://github.com/tensorflow/models/blob/master/research/slim/nets/mobilenet/mobilenet.py
    """
    n = v / divisor
    return int(max(n + 0.5, n * 0.9 + 1)) * divisor



__all__ = ['export', 'get_wild_imports', 'import_tree']

import pkgutil
import sys
from collections.abc import Callable
from types import ModuleType
from typing import TypeVar

_T = TypeVar('_T', bound=Callable)


class ExportError(Exception):
    pass


def export(obj: _T) -> _T:
    """Exposes obj to __all__ in module parent to where it was defined,
    although breaking intellisense.

    Usage:

    In package/subpackage.py:
    ```
    from glow.api import export

    @export
    def func():
        pass
    ```
    In package/__init__.py:
    ```
    from . import subpackage
    ```
    In main.py:
    ```
    # not `from package.subpackage import func`
    from package import func
    func()
    ```
    """
    spec = sys.modules[obj.__module__].__spec__
    if not spec or not spec.parent:
        return obj

    parent: str = spec.parent
    if obj.__module__ == parent:
        return obj

    name = obj.__name__
    namespace = sys.modules[parent].__dict__

    __all__ = namespace.setdefault('__all__', [])
    if name in __all__ or name in namespace:
        err = (f'Name "{name}" is reserved in <{parent}>'
               f' by <{namespace[name].__module__}:{name}>')
        raise ExportError(err)

    namespace[name] = obj
    __all__.append(name)
    return obj


def import_tree(pkg: str):
    """Imports all subpackages.

    Usage:
    ```
    __import__('glow.api', fromlist=['api']).import_tree(__name__)
    ```
    """
    path: str = sys.modules[pkg].__path__  # type: ignore
    for _, name, __ in pkgutil.walk_packages(path):
        subpkg = pkg + '.' + name
        __import__(subpkg)


def get_wild_imports(module: ModuleType) -> tuple[str, ...]:
    """Get contents of module.__all__ if possible"""
    __all__ = getattr(module, '__all__', None)
    if __all__ is not None:
        return __all__
    return *(name for name in dir(module) if not name.startswith('_')),



from __future__ import annotations

__all__ = ['conv2d_ws', 'upscale2d']

from typing import Union

import torch
import torch.nn.functional as F
from torch import nn
from torch.nn.utils import parametrize

_size = Union[torch.Size, list[int], tuple[int, ...]]


# @torch.jit.script
def upscale2d(x: torch.Tensor, stride: int = 2) -> torch.Tensor:
    # ! stride-aware fallback, works everywhere
    # x = F.interpolate(x, None, self.stride)
    # return F.avg_pool2d(x, self.stride, 1, 0)

    # ! matches to single libtorch op, complex torchscript op
    pad = 1 - stride
    size = [x.shape[2], x.shape[3]]
    size = [s * stride + pad for s in size]
    return F.interpolate(x, size, mode='bilinear', align_corners=True)


def conv2d_ws(x: torch.Tensor,
              weight: torch.Tensor,
              bias: torch.Tensor | None = None,
              stride: _size | int = 1,
              padding: _size | int | str = 0,
              dilation: _size | int = 1,
              groups: int = 1):
    weight = F.layer_norm(weight, weight.shape[1:], eps=1e-5)
    return F.conv2d(x, weight, bias, stride, padding, dilation, groups)


def standartize_conv_weights(model: nn.Module) -> None:
    """Enforce weight standartization for all ConvNd modules"""
    for m in model.modules():
        if not isinstance(m, nn.modules.conv._ConvNd):
            continue
        shape = m.weight.shape
        parametrize.register_parametrization(
            m, 'weight', nn.LayerNorm(shape[1:], elementwise_affine=False))



from __future__ import annotations

__all__ = ['affine', 'flip', 'grid_shuffle', 'mask_dropout']

import cv2
import numpy as np


def grid_shuffle(image: np.ndarray,
                 mask: np.ndarray | None = None,
                 *,
                 rng: np.random.Generator,
                 grid=4) -> tuple[np.ndarray, ...]:
    axes = (
        np.linspace(0, size, grid + 1, dtype=np.int_)
        for size in image.shape[:2])
    mats = np.stack(np.meshgrid(*axes, indexing='ij'))

    anchors = mats[:, :-1, :-1]
    tiles_sizes = (mats[:, 1:, 1:] - anchors).transpose(1, 2, 0)
    indices = np.indices((grid, grid)).transpose(1, 2, 0)

    for axis in range(2):
        indices = rng.permutation(indices, axis=axis)

    for bbox_size in np.unique(tiles_sizes.reshape(-1, 2), axis=0):
        eq_mat = np.all(tiles_sizes == bbox_size, axis=2)
        indices[eq_mat] = rng.permutation(indices[eq_mat])

    old_pos = anchors[:, indices[..., 0], indices[..., 1]].reshape(2, -1)
    new_pos = anchors.reshape(2, -1)

    sizes = tiles_sizes.reshape(-1, 2)
    tiles = np.stack([new_pos.T, old_pos.T, sizes], axis=1)

    sources = [image]
    if mask is not None:
        sources.append(mask)

    results = []
    for v in sources:
        new_v = np.empty_like(v)
        for (y2, x2), (y1, x1), (ys, xs) in tiles:
            new_v[y2:y2 + ys, x2:x2 + xs] = v[y1:y1 + ys, x1:x1 + xs]
        results.append(new_v)
    return *results,


def affine(image: np.ndarray,
           skew: float,
           angle: float,
           scale: float,
           inter: int = cv2.INTER_LINEAR,
           border: int = cv2.BORDER_REFLECT_101) -> np.ndarray:
    center = np.array(image.shape[1::-1]) / 2
    mat = np.hstack([np.eye(2), -center[:, None]])  # move center to origin

    angle = np.radians(angle)
    rotate = [[np.cos(angle), -np.sin(angle)], [np.sin(angle), np.cos(angle)]]

    shear = [[np.cosh(skew), np.sinh(skew)], [np.sinh(skew), np.cosh(skew)]]

    mat = np.array(rotate) @ np.array(shear) @ mat / scale

    mat[:, 2] += center  # restore center

    flags = inter + cv2.WARP_INVERSE_MAP
    return cv2.warpAffine(
        image, mat, image.shape[:2], flags=flags, borderMode=border)


def flip(image: np.ndarray, *, ud: bool, lr: bool, rot90: bool) -> np.ndarray:
    if ud:
        image = image[::-1, :]
    if lr:
        image = image[:, ::-1]
    if rot90:
        image = np.rot90(image, axes=(0, 1))
    return np.ascontiguousarray(image)


def mask_dropout(mask: np.ndarray,
                 rng: np.random.Generator,
                 alpha: float,
                 ignore_index: int = -1) -> np.ndarray:
    """
    Keep occurence of each class below alpha.
    Redundant values are replaced with ignore_index.
    """
    dist = np.bincount(mask.ravel()) / mask.size
    prob_keep = alpha / np.maximum(alpha, dist)

    mask = mask.astype('i8')
    mask[rng.random(mask.shape, dtype='f4') >= prob_keep[mask]] = ignore_index
    return mask


def gamma(arr: np.ndarray, y: float) -> np.ndarray:
    """Apply gamma Y to array"""
    assert arr.dtype == 'u1'
    assert y > 0

    lut = np.arange(256, dtype='f4')
    lut /= 255
    lut **= y
    lut *= 255
    lut = lut.round().astype('u1')

    return cv2.LUT(arr.ravel(), lut).reshape(arr.shape)


def gamma_dq(arr: np.ndarray,
             y: float,
             rng: np.random.Generator,
             qbits: int = 7) -> np.ndarray:
    """Apply gamma Y to array with dequantization"""
    assert arr.dtype == 'u1'
    assert y > 0
    assert 0 <= qbits < 8
    peak = 256 << qbits

    lut = np.linspace(0, 1, num=257, dtype='f4')
    lut **= y
    lut *= peak  # Increase precision
    lut = lut.astype('u2')
    lut = lut.clip(0, peak - 1)  # Remove overshoot

    arr = rng.integers(lut[arr], lut[1:][arr], dtype='u2', endpoint=True)
    arr >>= qbits  # Round to 8 bits
    return arr.astype('u1')



from collections.abc import Callable, Iterable
from contextlib import AbstractContextManager
from typing import TypeVar, overload

from typing_extensions import ParamSpec

_T = TypeVar('_T')
_F = TypeVar('_F', bound=Callable)
_P = ParamSpec('_P')
_BatchedFn = TypeVar('_BatchedFn', bound=Callable[[list], Iterable])


def threadlocal(fn: Callable[_P, _T], *args: _P.args,
                **kwargs: _P.kwargs) -> Callable[[], _T]:
    ...


def interpreter_lock(timeout: float = ...) -> AbstractContextManager[None]:
    ...


def call_once(fn: _F, /) -> _F:
    ...


def shared_call(fn: _F, /) -> _F:
    ...


def weak_memoize(fn: _F, /) -> _F:
    ...


@overload
def streaming(*,
              batch_size: int,
              timeout: float = ...,
              workers: int = ...,
              pool_timeout: float = ...) -> Callable[[_BatchedFn], _BatchedFn]:
    ...


@overload
def streaming(func: _BatchedFn,
              *,
              batch_size: int,
              timeout: float = ...,
              workers: int = ...,
              pool_timeout: float = ...) -> _BatchedFn:
    ...



from . import _patch_len, _patch_print, _patch_scipy
from ._coro import as_actor, coroutine, summary
from ._more import as_iter, chunked, eat, ichunked, ilen, roundrobin, windowed
from ._parallel import buffered, map_n, map_n_dict, starmap_n
from ._profile import memprof, time_this, timer
from ._repr import countable, mangle, repr_as_obj, si, si_bin
from ._sizeof import sizeof
from ._uuid import Uid
from .debug import lock_seed, trace, trace_module, whereami
from .wrap import (Reusable, call_once, memoize, shared_call, streaming,
                   threadlocal, weak_memoize)

__all__ = [
    'Reusable', 'Uid', 'as_actor', 'as_iter', 'buffered', 'call_once',
    'chunked', 'coroutine', 'countable', 'eat', 'ichunked', 'ilen',
    'lock_seed', 'mangle', 'map_n', 'map_n_dict', 'memoize', 'memprof',
    'repr_as_obj', 'roundrobin', 'shared_call', 'si', 'si_bin', 'sizeof',
    'starmap_n', 'streaming', 'summary', 'threadlocal', 'time_this', 'timer',
    'trace', 'trace_module', 'weak_memoize', 'whereami', 'windowed'
]

_patch_print.apply()
_patch_len.apply()
_patch_scipy.apply()



"""argparse and dataclasses, married.

Example:
```
@dataclass
class Args:
    a: int
    b: str = 'hello'

args, parser = parse_args(Args)
```
Or with plain function:
```
@parse_args
def main(name: str = 'user'):
    print(f'Hello {name}')
```

Reasons not to use alternatives:
- [simple-parsing](https://github.com/lebrice/SimpleParsing):
  - Has underscores (`--like_this`) instead of dashes (`--like-this`)
  - Erases type on parser result, thus making typed prototype useless
    (what is the point of using dataclasses if it is passed
     through function returning typing.Any/argparse.Namespace?)

- [datargs](https://github.com/roee30/datargs):
  - No nesting support
  - No function's support

- [pydantic](https://github.com/samuelcolvin/pydantic):
  - supports CLI via BaseSettings and environment variables parsing
  - no nesting, as requires mixing only via multiple inheritance

- [typer](https://github.com/tiangolo/typer):
  - No support on dataclasses
    (https://github.com/tiangolo/typer/issues/154).
  - No fine way to extract parsed options without invoking because of
    decorator/callback based implementation. Thus enforces wrapping of the
    whole app into `typer.run`.
    (https://github.com/tiangolo/typer/issues/197).
"""

from __future__ import annotations

__all__ = ['arg', 'parse_args']

import argparse
import sys
import types
from argparse import ArgumentParser, BooleanOptionalAction, _ArgumentGroup
from collections.abc import Callable, Collection, Iterator, Sequence
from dataclasses import MISSING, Field, field, fields, is_dataclass
from inspect import getmodule, signature, stack
from typing import (Any, Literal, TypeVar, Union, get_args, get_origin,
                    get_type_hints)

_T = TypeVar('_T')
_Node = Union[str, tuple[str, type, list['_Node']]]
_NoneType = type(None)
_UNION_TYPES: list = [Union]

if sys.version_info >= (3, 10):
    _UNION_TYPES += [types.UnionType]


def arg(
    default=MISSING,
    /,
    *,
    flag=None,
    factory=MISSING,
    init=True,
    repr=True,  # noqa: A002
    hash=None,  # noqa: A002
    help=None,  # noqa: A002
    compare=True,
    metadata=None,
):
    """Convinient alias for dataclass.field with extra metadata (like help)"""
    metadata = metadata or {}
    for k, v in {'flag': flag, 'help': help}.items():
        if v:
            metadata = metadata | {k: v}
    return field(  # type: ignore[call-overload]
        default=default,
        default_factory=factory,
        init=init,
        repr=repr,
        hash=hash,
        compare=compare,
        metadata=metadata)


def _unwrap_type(tp: type) -> tuple[type, dict[str, Any]]:
    if tp is list:
        raise ValueError('Type list should be parametrized')

    origin = get_origin(tp)
    *args, = get_args(tp)
    if not origin or not args:  # Not a generic type
        return tp, {'type': tp}

    if origin is list:  # `List[T]`
        cls, opts = _unwrap_type(args[0])
        return cls, opts | {'nargs': argparse.ZERO_OR_MORE}

    # `Optional[T]` or `T | None`
    if origin in _UNION_TYPES and len(args) == 2 and _NoneType in args:
        args.remove(_NoneType)
        cls, opts = _unwrap_type(args[0])
        if opts.get('nargs') == argparse.ZERO_OR_MORE:
            return cls, opts
        return cls, opts | {'nargs': argparse.OPTIONAL}

    if origin is Literal:  # `Literal[x, y]`
        choices = get_args(tp)
        if len(tps := {type(c) for c in choices}) != 1:
            raise ValueError('Literal parameters should have '
                             f'the same type. Got: {tps}')
        cls, = tps
        return cls, {'type': cls, 'choices': choices}

    raise ValueError('Only list, Optional and Literal are supported '
                     f'as generic types. Got: {tp}')


def _get_fields(fn: Callable) -> Iterator[Field]:
    if is_dataclass(fn):  # Shortcut
        yield from fields(fn)
        return

    for p in signature(fn).parameters.values():
        if p.kind is p.KEYWORD_ONLY and p.default is p.empty:
            raise ValueError(f'Keyword "{p.name}" must have default')
        if p.kind in {p.POSITIONAL_ONLY, p.VAR_POSITIONAL, p.VAR_KEYWORD}:
            raise ValueError(f'Unsupported parameter type: {p.kind}')

        if isinstance(p.default, Field):
            fd = p.default
        else:
            fd = arg(MISSING if p.default is p.empty else p.default)
        fd.name = p.name
        yield fd


def _visit_nested(parser: ArgumentParser | _ArgumentGroup, fn: Callable,
                  seen: dict[str, list]) -> list[_Node]:
    try:
        hints = get_type_hints(fn)
    except NameError:
        if fn.__module__ != '__main__':
            raise
        for finfo in stack():
            if not getmodule(finfo.frame):
                hints = get_type_hints(fn, finfo.frame.f_globals)
                break
        else:
            raise

    nodes: list[_Node] = []
    for fd in _get_fields(fn):
        if fd.init:
            seen.setdefault(fd.name, []).append(fn)
            nodes.append(_visit_field(parser, hints[fd.name], fd, seen))

    for name, usages in seen.items():
        if len(usages) > 1:
            raise ValueError(f'Field name "{name}" occured multiple times: '
                             + ', '.join(f'{c.__module__}.{c.__qualname__}'
                                         for c in usages)
                             + '. All field names should be unique')
    return nodes


def _visit_field(parser: ArgumentParser | _ArgumentGroup, tp: type, fd: Field,
                 seen: dict[str, list]) -> _Node:
    cls, opts = _unwrap_type(tp)

    help_ = fd.metadata.get('help') or ''
    if cls is not bool and fd.default is not MISSING:
        help_ += f' (default: {fd.default})'

    if is_dataclass(cls):  # Nested dataclass
        arg_group = parser.add_argument_group(fd.name)
        return fd.name, cls, _visit_nested(arg_group, cls, seen)

    snake = fd.name.replace('_', '-')
    flags = [f] if (f := fd.metadata.get('flag')) else []

    if cls is bool:  # Optional
        if fd.default is MISSING:
            raise ValueError(f'Boolean field "{fd.name}" should have default')
        parser.add_argument(
            f'--{snake}',
            *flags,
            action=BooleanOptionalAction,
            default=fd.default,
            help=help_)

    # Generic optional
    elif fd.default is not MISSING or fd.default_factory is not MISSING:
        if opts.get('nargs') == argparse.OPTIONAL:
            del opts['nargs']
        default = fd.default_factory() if fd.default is MISSING else fd.default
        parser.add_argument(
            f'--{snake}', *flags, **opts, default=default, help=help_)

    elif isinstance(parser, ArgumentParser):  # Allow only for root parser
        if flags:
            raise ValueError(f'Positional-only field "{fd.name}" '
                             'should not have flag')
        parser.add_argument(snake, **opts, help=help_)

    else:
        raise ValueError('Positional-only fields are forbidden '
                         'for nested types. Please set default value '
                         f'for "{fd.name}"')

    return fd.name


def _construct(src: dict[str, Any], fn: Callable[..., _T],
               args: Collection[_Node]) -> _T:
    kwargs = {}
    for a in args:
        if isinstance(a, str):
            kwargs[a] = src.pop(a)
        else:
            kwargs[a[0]] = _construct(src, a[1], a[2])
    return fn(**kwargs)


def parse_args(fn: Callable[..., _T],
               args: Sequence[str] | None = None,
               prog: str | None = None) -> tuple[_T, ArgumentParser]:
    """Create parser from type hints of callable, parse args and do call"""
    # TODO: Rename to `run`
    parser = ArgumentParser(prog)
    nodes = _visit_nested(parser, fn, {})

    if args is not None:  # fool's protection
        args = line.split(' ') if (line := ' '.join(args).strip()) else []

    namespace = parser.parse_args(args)
    obj = _construct(vars(namespace), fn, nodes)
    return obj, parser



from collections.abc import Callable, Hashable, Iterable
from typing import Literal, TypeAlias, TypeVar, overload

_F = TypeVar('_F', bound=Callable)
_BatchedFn = TypeVar('_BatchedFn', bound=Callable[[list], Iterable])
_Policy: TypeAlias = Literal['raw', 'lru', 'mru']
_KeyFn: TypeAlias = Callable[..., Hashable]


@overload
def memoize(capacity: int,
            *,
            policy: _Policy = ...,
            key_fn: _KeyFn = ...,
            bytesize: bool = ...) -> Callable[[_F], _F]:
    ...


@overload
def memoize(capacity: int,
            *,
            batched: Literal[True],
            policy: _Policy = ...,
            key_fn: _KeyFn = ...,
            bytesize: bool = ...) -> Callable[[_BatchedFn], _BatchedFn]:
    ...



from importlib import import_module
from typing import TYPE_CHECKING

from ._loader import get_loader
from ._trainer import Trainer
from .amp import get_grads
from .modules import (Ensemble, Noise, Residual, SplitAttention,
                      SqueezeExcitation, Upscale2d)
from .optimizers import SGDW, AdamW, RAdam
from .util import (detach_, device, dump_to_onnx, eval_, frozen, inference,
                   param_count, profile)

__all__ = [
    'AdamW', 'Ensemble', 'Noise', 'RAdam', 'Residual', 'SGDW',
    'SplitAttention', 'SqueezeExcitation', 'Trainer', 'Upscale2d', 'detach_',
    'device', 'dump_to_onnx', 'eval_', 'frozen', 'get_gpu_capability',
    'get_gpu_memory_info', 'get_grads', 'get_loader', 'inference',
    'param_count', 'plot_model', 'profile'
]

_exports = {
    '.plot': ['plot_model'],
    '.driver': ['get_gpu_capability', 'get_gpu_memory_info'],
}
_submodule_by_name = {
    name: modname for modname, names in _exports.items() for name in names
}

if TYPE_CHECKING:
    from .driver import get_gpu_capability, get_gpu_memory_info
    from .plot import plot_model

else:

    def __getattr__(name: str):
        if modname := _submodule_by_name.get(name):
            mod = import_module(modname, __package__)
            globals()[name] = obj = getattr(mod, name)
            return obj
        raise AttributeError(f'No attribute {name}')

    def __dir__():
        return __all__



from argparse import ArgumentParser
from collections.abc import Callable, Mapping, Sequence
from typing import Any, TypeVar, overload

_T = TypeVar('_T')


@overload
def arg(
        default: _T,
        /,
        *,
        flag: str = ...,
        init: bool = ...,
        repr: bool = ...,  # noqa: A002
        hash: bool = ...,  # noqa: A002
        help: str = ...,  # noqa: A002
        compare: bool = ...,
        metadata: Mapping[str, object] = ...) -> _T:
    ...


@overload
def arg(
        *,
        factory: Callable[[], _T],
        flag: str = ...,
        init: bool = ...,
        repr: bool = ...,  # noqa: A002
        hash: bool = ...,  # noqa: A002
        help: str = ...,  # noqa: A002
        compare: bool = ...,
        metadata: Mapping[str, object] = ...) -> _T:
    ...


@overload
def arg(
        *,
        flag: str = ...,
        init: bool = ...,
        repr: bool = ...,  # noqa: A002
        hash: bool = ...,  # noqa: A002
        help: str = ...,  # noqa: A002
        compare: bool = ...,
        metadata: Mapping[str, object] = ...) -> Any:
    ...


def parse_args(fn: Callable[..., _T],
               args: Sequence[str] = ...,
               prog: str = ...) -> tuple[_T, ArgumentParser]:
    ...



import time
from dataclasses import dataclass
from itertools import count

import glow


@dataclass
class Value:
    value: int


def test_base():
    counter = count()
    ref = glow.Reusable(lambda: Value(next(counter)), delay=0.05)
    assert ref


def test_fail():
    counter = count()
    ref = glow.Reusable(lambda: Value(next(counter)), delay=0.05)
    assert ref().value == 0

    time.sleep(0.1)
    assert ref().value == 1


def test_success() -> None:
    counter = count()
    ref = glow.Reusable(lambda: Value(next(counter)), delay=0.1)
    assert ref().value == 0

    time.sleep(0.05)
    assert ref().value == 0


def test_success_double():
    counter = count()
    ref = glow.Reusable(lambda: Value(next(counter)), delay=0.1)
    assert ref().value == 0

    time.sleep(0.05)
    assert ref().value == 0

    time.sleep(0.05)
    assert ref().value == 0



from .cache import memoize
from .concurrency import (call_once, shared_call, streaming, threadlocal,
                          weak_memoize)
from .reusable import Reusable

__all__ = [
    'Reusable', 'call_once', 'memoize', 'shared_call', 'streaming',
    'threadlocal', 'weak_memoize'
]



from __future__ import annotations

__all__ = ['get_loader']

import os
import re
import warnings
from collections.abc import Callable, Iterator, Mapping, Sequence, Sized
from dataclasses import dataclass, replace
from functools import partial
from itertools import islice
from typing import Any, Protocol

import numpy as np
import torch
from torch.utils.data import (Dataset, IterableDataset, RandomSampler, Sampler,
                              SequentialSampler)
from torch.utils.data._utils import worker as torch_worker

from .. import buffered, chunked, map_n, roundrobin
from ..core._parallel import _get_executor, max_cpu_count
from ..distributed import get_ddp_info
from ._sampler import DdpSampler, SamplerLike, generate_seed
from .util import _apply

_NUM_CPUS: int = os.cpu_count() or 1

# ------------------------------- common bases -------------------------------


class _Loader(Protocol):
    def __iter__(self) -> Iterator:
        raise NotImplementedError

    def __len__(self) -> int:
        raise NotImplementedError


# ------------------------------ memory pinning ------------------------------


def pin_memory(data):
    def _pin_memory(x):
        return x.pin_memory()

    return _apply(data, _pin_memory)


@dataclass(frozen=True)
class _PinningLoader(_Loader):
    base: _Loader

    def __iter__(self) -> Iterator:
        return map_n(pin_memory, self.base, max_workers=1)

    def __len__(self) -> int:
        return len(self.base)


class _PinnableLoader(_Loader):
    def pin(self) -> _Loader:
        """
        Copies Tensors into device/CUDA pinned memory before returning them.
        Works in extra thread.
        """
        return _PinningLoader(self) if torch.cuda.is_available() else self


# --------------------------------- batching ---------------------------------


class _CollateFn(Protocol):
    def __call__(self, __items: tuple) -> Any:
        ...


@dataclass(frozen=True)
class _BatchedLoader(_PinnableLoader):
    base: _Loader
    batch_size: int
    collate_fn: _CollateFn
    drop_last: bool
    workers: int

    def __iter__(self) -> Iterator:
        it = iter(self.base)

        if self.drop_last:
            total = len(self) * self.batch_size
            it = islice(it, total)

        batches = chunked(it, self.batch_size)
        return map_n(self.collate_fn, batches, max_workers=self.workers)

    def __len__(self) -> int:
        len_ = len(self.base)

        if self.drop_last:
            return len_ // self.batch_size

        return len(range(0, len_, self.batch_size))


class _BatchableLoader(_PinnableLoader):
    def batch(self,
              batch_size: int,
              collate_fn: _CollateFn | None = None,
              drop_last: bool = False,
              daemon: bool = False) -> _BatchedLoader:
        """
        Groups data into batches.

        Parameters:
        - batch_size - How many samples per batch to load
        - collate_fn - Merges a list of samples to form a mini-batch of
          Tensor(s).
        - drop_last - Set to `True` to drop the last incomplete batch,
          if size of underlying loader is not divisible by the batch size.
          Otherwise the last batch can be smaller than others.
        - daemon - Set to do batching in background thread.
        """
        if collate_fn is None:
            collate_fn = collate
        # TODO: return _BatchedLoader(
        #   replace(self, finalize=None),
        #   batch_size, collate_fn, drop_last,
        # )
        return _BatchedLoader(self, batch_size, collate_fn, drop_last,
                              int(daemon))

    def shuffle(self,
                sampler: Sampler | SamplerLike | None) -> _BatchableLoader:
        """
        Reshuffle data at every epoch.

        Parameters:
        - sampler - Defines the strategy to draw samples from the dataset.
          Can be any `Iterable` with `__len__` implemented.
        """
        raise NotImplementedError


# ---------------------- loader for map-style datasets ----------------------


@dataclass(frozen=True)
class _MapLoader(_BatchableLoader):
    dataset: Dataset
    sampler: Sampler

    def __iter__(self) -> Iterator:
        return map(self.dataset.__getitem__, self.sampler)

    def __len__(self) -> int:
        assert isinstance(self.sampler, Sized)
        return len(self.sampler)

    def shuffle(self,
                sampler: Sampler | SamplerLike | bool | None) -> _MapLoader:
        if not sampler:
            return self

        if sampler is True:
            assert isinstance(self.dataset, Sized)
            sampler = RandomSampler(self.dataset)

        return replace(self, sampler=DdpSampler(sampler))


@dataclass(frozen=True)
class _MapMultiLoader(_MapLoader):
    max_workers: int
    mp: bool
    chunksize: int | None = None

    def __iter__(self) -> Iterator:
        max_workers = self.max_workers
        if ddp := get_ddp_info():
            max_workers //= ddp.world

        return map_n(
            self.dataset.__getitem__,
            self.sampler,
            max_workers=max_workers,
            chunksize=self.chunksize,
            mp=self.mp)


# -------------------- loader for iterable-style datasets --------------------


@dataclass(frozen=True)
class _Worker:
    dataset: IterableDataset
    id: int
    num_workers: int
    seed: int | None = None

    def __iter__(self) -> Iterator:
        torch_worker._worker_info = self  # type: ignore[assignment]
        try:
            yield from self.dataset
        finally:
            torch_worker._worker_info = None


@dataclass(frozen=True)
class _IterableLoader(_BatchableLoader):
    dataset: IterableDataset

    def __iter__(self) -> Iterator:
        return buffered(self.dataset)

    def __len__(self) -> int:
        assert isinstance(self.dataset, Sized)
        return len(self.dataset)


@dataclass(frozen=True)
class _IterableMultiLoader(_IterableLoader):
    max_workers: int

    def __iter__(self) -> Iterator:
        seed = generate_seed()
        workers = [
            _Worker(self.dataset, idx, self.max_workers, seed)
            for idx in range(self.max_workers)
        ]
        with _get_executor(self.max_workers, True) as ex:
            yield from roundrobin(*(buffered(w, mp=ex) for w in workers))


# ----------------------------- factory function -----------------------------


def get_loader(dataset: Dataset,
               max_workers: int | None = 0,
               mp: bool = False) -> _BatchableLoader:
    """
    Data loader. Combines a dataset and a sampler (via shuffle method), and
    provides an iterable over the given dataset.

    The data loader supports both map-style and iterable-style datasets with
    single- or multi-process loading, customizing loading order and
    automatic batching (collation) and memory pinning.

    Differences from torch.utils.data.DataLoader:
    - Support of threadpool backend for map-style datasets.
    - Automatically adjusts chunksize of data for passing to the workers to
      reduce IPC overhead when multiple processes are used.
    - Automatically adjusts number of workers in distributed context for
      map-style datasets.

    Parameters:
    - dataset - Dataset from which to load the data.
    - max_workers - How many threads or subprocesses to use for data loading.
      `0` means that the data will be loaded in the main process/thread.
      `None` means all logical processors.
    - mp - Whether to use multiprocessing or not.
    """
    if max_workers is None:
        max_workers = max_cpu_count(_NUM_CPUS, mp)

    if isinstance(dataset, IterableDataset):
        if not mp and max_workers != 0:
            warnings.warn(
                'For iterable-style datasets multithreading is not supported. '
                'Setting max_workers to 0',
                stacklevel=2)
            max_workers = 0

        if (ddp := get_ddp_info()) and ddp.world > 1:
            raise ValueError(
                'For iterable-style datasets distributed use is not '
                'supported')

        if not max_workers:
            return _IterableLoader(dataset)
        return _IterableMultiLoader(dataset, max_workers)

    else:  # noqa; RET505
        if not isinstance(dataset, Sized):
            raise TypeError("dataset should be sized when it's not iterable")

        sampler = SequentialSampler(dataset)

        if not max_workers:
            return _MapLoader(dataset, DdpSampler(sampler))
        return _MapMultiLoader(dataset, DdpSampler(sampler), max_workers, mp)


# ---------------- convert & collate. forked from pytorch 2.0 ----------------

# TODO: split `collate` to `convert` (within worker) + `collate` (in main)
# TODO: i.e. "Loader".tensors [alters workers] .batch() [adds collation]


def convert(x):  # noqa: PLR0911
    tp = type(x)
    if isinstance(x, torch.Tensor):
        return x

    if tp.__module__ == 'numpy' and not isinstance(x, np.str_ | np.bytes_):
        if (isinstance(x, np.ndarray)
                and _NP_STR_DTYPE_PATTERN.search(x.dtype.str)):
            return x
        return torch.as_tensor(x)

    if isinstance(x, Mapping):
        return _apply_type(tp, {k: convert(v) for k, v in x.items()})

    if isinstance(x, Sequence) and not isinstance(x, str | bytes):
        list_ = [convert(xx) for xx in x]

        if isinstance(x, tuple) and hasattr(x, '_fields'):  # namedtuple
            return tp(*list_)
        return _apply_type(tp, list_)

    return x


def collate(batch):
    x0 = batch[0]
    tp = type(x0)

    if _HINT is not None:  # Fast alternative to functools.singledispatch
        fn = _HINT.get(tp) or next(
            (fn for tp_, fn in _HINT.items() if isinstance(x0, tp_)), None)
        if fn is not None:
            return fn(batch)

    if isinstance(x0, Mapping):
        return _apply_type(tp, {k: collate([x[k] for x in batch]) for k in x0})

    if isinstance(x0, Sequence):
        assert len({len(x) for x in batch}) <= 1  # py3.10+: zip(strict=True)
        list_ = [collate(samples) for samples in zip(*batch)]

        if isinstance(x0, tuple) and hasattr(x0, '_fields'):  # namedtuple
            return tp(*list_)
        return _apply_type(tp, list_)

    raise TypeError(_COLLATE_ERROR_MSG.format(tp))


def _apply_type(tp, x):
    try:
        return tp(x)
    except TypeError:
        return x


def _collate_tensor(batch: Sequence[torch.Tensor]):
    x = batch[0]
    out = None
    if torch_worker.get_worker_info() is not None:
        # If we're in a background process, concatenate directly into a
        # shared memory tensor to avoid an extra copy
        numel = sum(x.numel() for x in batch)
        storage = x._typed_storage()._new_shared(numel, device=x.device)
        out = x.new(storage).resize_(len(batch), *x.shape)

    return torch.stack([*batch], out=out)


def _collate_ndarray(batch: Sequence[np.ndarray]):
    x = batch[0]
    if _NP_STR_DTYPE_PATTERN.search(x.dtype.str):
        raise TypeError(_COLLATE_ERROR_MSG.format(x.dtype))

    return collate([torch.as_tensor(x) for x in batch])


def _nop(batch):
    return batch


_HINT: dict[type | tuple[type, ...], Callable] = {
    torch.Tensor: _collate_tensor,
    np.ndarray: _collate_ndarray,  # For both ndarray and memmap
    # Skip strings
    bytes: _nop,
    str: _nop,
    # Tensorify scalars
    (np.bool_, np.number, np.object_): torch.as_tensor,  # py3.10: UnionType
    float: partial(torch.tensor, dtype=torch.float64),
    int: torch.tensor,
}

_NP_STR_DTYPE_PATTERN = re.compile('[SOUa]')
_COLLATE_ERROR_MSG = (
    'default_collate: batch must contain tensors, numpy arrays, numbers, '
    'dicts or lists; found {}')



__all__ = [
    'LazyBias2d', 'LazyBlurPool2d', 'LazyConv2dWs', 'LazyGroupNorm',
    'LazyLayerNorm'
]

import warnings

import torch
from torch import nn
from torch.nn.parameter import UninitializedBuffer, UninitializedParameter

from .simple import Bias2d, BlurPool2d, Conv2dWs

warnings.filterwarnings('ignore', module='torch.nn.modules.lazy')


def _materialize_cls(m: nn.modules.lazy._LazyProtocol):
    # Fixes incomplete implementation of LazyModuleMixin._lazy_load_hook

    # Does the same as LazyModuleMixin._infer_parameters does,
    # except no input is needed

    # By default, if all module's parameters are loaded during load_state_dict,
    # _lazy_load_hook doen't mutate class.
    # Because of that even completely initialized modules are left as lazy,
    # and require calling forward() to trigger class mutation.

    # This function cancels this requirement.

    # FIXME: When merged to an upstream, do nothing
    assert isinstance(m, nn.modules.lazy.LazyModuleMixin)

    m._initialize_hook.remove()
    m._load_hook.remove()
    delattr(m, '_initialize_hook')
    delattr(m, '_load_hook')
    if m.cls_to_become is not None:
        m.__class__ = m.cls_to_become


class _LazyModuleMixinV2(nn.modules.lazy.LazyModuleMixin):
    def _lazy_load_hook(self: nn.modules.lazy._LazyProtocol, *args, **kwargs):
        super()._lazy_load_hook(*args, **kwargs)  # type: ignore

        if not self.has_uninitialized_params():  # type: ignore
            _materialize_cls(self)


class _LazyBase(_LazyModuleMixinV2):
    def reset_parameters(self) -> None:
        if not self.has_uninitialized_params():  # type: ignore
            super().reset_parameters()  # type: ignore

    def initialize_parameters(self, x: torch.Tensor) -> None:
        if self.has_uninitialized_params():  # type: ignore
            self.materialize(x.shape)
            self.reset_parameters()

    def materialize(self, shape: torch.Size) -> None:
        raise NotImplementedError


class LazyLayerNorm(_LazyBase, nn.LayerNorm):
    cls_to_become = nn.LayerNorm  # type: ignore

    weight: UninitializedParameter  # type: ignore
    bias: UninitializedParameter  # type: ignore

    def __init__(self,
                 rank: int = 1,
                 eps: float = 1e-5,
                 elementwise_affine: bool = True):
        super().__init__([0] * rank, eps, False)
        self.elementwise_affine = elementwise_affine
        if self.elementwise_affine:
            self.weight = UninitializedParameter()
            self.bias = UninitializedParameter()

    def materialize(self, shape: torch.Size) -> None:
        rank = len(self.normalized_shape)
        self.normalized_shape = *shape[-rank:],
        if self.elementwise_affine:
            self.weight.materialize(self.normalized_shape)
            self.bias.materialize(self.normalized_shape)


class LazyGroupNorm(_LazyBase, nn.GroupNorm):
    cls_to_become = nn.GroupNorm  # type: ignore

    weight: UninitializedParameter  # type: ignore
    bias: UninitializedParameter  # type: ignore

    def __init__(self,
                 num_groups: int,
                 eps: float = 1e-5,
                 affine: bool = True):
        super().__init__(num_groups, 0, eps, False)
        self.affine = affine
        if self.affine:
            self.weight = UninitializedParameter()
            self.bias = UninitializedParameter()

    def materialize(self, shape: torch.Size) -> None:
        self.num_channels = shape[1]
        if self.affine:
            self.weight.materialize((self.num_channels, ))
            self.bias.materialize((self.num_channels, ))


class LazyConv2dWs(nn.LazyConv2d):
    cls_to_become = Conv2dWs  # type: ignore


class LazyBlurPool2d(_LazyBase, BlurPool2d):
    cls_to_become = BlurPool2d  # type: ignore

    weight: UninitializedBuffer

    def __init__(self,
                 kernel: int = 4,
                 stride: int = 2,
                 padding: int = 1,
                 padding_mode: str = 'reflect'):
        super().__init__(0, kernel, stride, padding, padding_mode)
        self.weight = UninitializedBuffer()

    def materialize(self, shape: torch.Size) -> None:
        self.in_channels = self.out_channels = shape[-3]
        self.weight.materialize((self.out_channels, 1, *self.kernel_size))


class LazyBias2d(_LazyBase, Bias2d):
    cls_to_become = Bias2d  # type: ignore

    bias: UninitializedParameter  # type: ignore

    def __init__(self):
        super().__init__(0, 0, 0)
        self.bias = UninitializedParameter()

    def materialize(self, shape: torch.Size) -> None:
        self.bias.materialize((1, *shape[1:]))



from __future__ import annotations

__all__ = ['memprof', 'time_this', 'timer']

import atexit
from collections import defaultdict, deque
from collections.abc import Callable, Iterator
from contextlib import contextmanager
from dataclasses import dataclass, field
from functools import partial
from itertools import accumulate, count
from threading import get_ident
from time import perf_counter_ns, process_time_ns, thread_time_ns
from typing import TYPE_CHECKING

from wrapt import ObjectProxy

from ._repr import si, si_bin
from .debug import whereami

if TYPE_CHECKING:
    import psutil
    _THIS: psutil.Process | None

_THIS = None


@contextmanager
def memprof(name_or_callback: str | Callable[[float], object] | None = None,
            /) -> Iterator[None]:
    global _THIS  # noqa: PLW0603
    if _THIS is None:
        import psutil
        _THIS = psutil.Process()

    init = _THIS.memory_info().rss
    try:
        yield
    finally:
        size = _THIS.memory_info().rss - init
        if callable(name_or_callback):
            name_or_callback(size)
        else:
            name = name_or_callback
            if name is None:
                name = f'{whereami(2, 1)} line'
            sign = '+' if size >= 0 else ''
            print(f'{name} done: {sign}{si_bin(size)}')


@contextmanager
def timer(name_or_callback: str | Callable[[int], object] | None = None,
          time: Callable[[], int] = perf_counter_ns,
          /) -> Iterator[None]:
    begin = time()
    try:
        yield
    finally:
        end = time()
        if callable(name_or_callback):
            name_or_callback(end - begin)
        else:
            name = name_or_callback or f'{whereami(2, 1)} line'
            print(f'{name} done in {si((end - begin) / 1e9)}s')


def _to_fname(obj) -> str:
    if not hasattr(obj, '__module__') or not hasattr(obj, '__qualname__'):
        obj = type(obj)
    if obj.__module__ == 'builtins':
        return obj.__qualname__
    return f'{obj.__module__}.{obj.__qualname__}'


class _Times(dict[int, int]):
    def add(self, value: int):
        idx = get_ident()
        self[idx] = self.get(idx, 0) + value

    def total(self) -> int:
        return sum(self.values())


class _Nlwp:
    __slots__ = ('_add_event', '_get_max')

    def __init__(self) -> None:
        events = deque[int]()
        self._add_event = events.append

        deltas = iter(events.popleft, None)
        totals = accumulate(deltas)
        maximums = accumulate(totals, max, initial=0)
        self._get_max = maximums.__next__

    def __enter__(self):
        self._add_event(+1)
        self._get_max()

    def __exit__(self, *args):
        self._add_event(-1)
        self._get_max()

    def max(self) -> int:
        self._add_event(0)
        return self._get_max()


@dataclass(frozen=True)
class _Stat:
    calls: count = field(default_factory=count)
    nlwp: _Nlwp = field(default_factory=_Nlwp)
    cpu_ns: _Times = field(default_factory=_Times)
    all_ns: _Times = field(default_factory=_Times)

    def __call__(self, op, *args, **kwargs):
        with self.nlwp, \
                timer(self.all_ns.add), \
                timer(self.cpu_ns.add, thread_time_ns):
            return op(*args, **kwargs)

    def stat(self) -> tuple[float, float, str] | None:
        if not (n := next(self.calls)):
            return None
        w = self.nlwp.max()
        t_ns = self.cpu_ns.total()  # CPU
        i_ns = self.all_ns.total() - t_ns  # idle = total - CPU
        t, i = t_ns / 1e9, i_ns / 1e9

        tail = (f'{n} x {si(t / n)}s' +
                (f' @ {w}T' if w > 1 else '')) if n > 1 else ''
        return t, i, tail


class _Proxy(ObjectProxy):
    def __init__(self, wrapped, wrapper):
        super().__init__(wrapped)
        self._self_wrapper = wrapper


class _TimedCall(_Proxy):
    def __get__(self, instance, owner):
        fn = self.__wrapped__.__get__(instance, owner)
        return _BoundTimedCall(fn, self._self_wrapper)

    def __call__(self, *args, **kwargs):
        next(self._self_wrapper.calls)
        r = self._self_wrapper(self.__wrapped__, *args, **kwargs)
        if isinstance(r, Iterator):
            return _TimedIter(r, self._self_wrapper)
        return r


class _BoundTimedCall(_TimedCall):
    def __get__(self, instance, owner):
        return self


class _TimedIter(_Proxy):
    def __iter__(self):
        return self

    def __next__(self):
        return self._self_wrapper(self.__wrapped__.__next__)

    def send(self, value):
        return self._self_wrapper(self.__wrapped__.send, value)

    def throw(self, typ, val=None, tb=None):
        return self._self_wrapper(self.__wrapped__.throw, typ, val, tb)

    def close(self):
        return self._self_wrapper(self.__wrapped__.close)


# Wall time, i.e. sum of per-thread times, excluding sleep
_start = process_time_ns()
_stats = defaultdict[str, _Stat](_Stat)


@atexit.register
def _print_stats(*names: str):
    all_busy = (process_time_ns() - _start + 1) / 1e9

    stats = []
    names = names or tuple(_stats)
    for name in names:
        if not (stat := _stats.pop(name, None)):
            continue
        if not (lines := stat.stat()):
            continue
        stats.append((*lines, name))

    for busy, idle, tail, name in sorted(stats):
        print(
            f'{busy/all_busy:6.2%}',
            f'{si(busy):>5s}s + {si(idle):>5s}s',
            name,
            tail,
            sep=' - ')


def time_this(fn=None, /, *, name: str | None = None):
    """Log function and/or generator timings at program exit"""
    if fn is None:
        return partial(time_this, name=name)

    if name is None:
        name = _to_fname(fn)

    fn.log_timing = partial(_print_stats, name)
    return _TimedCall(fn, _stats[name])



from __future__ import annotations

__all__ = ['buffered', 'map_n', 'map_n_dict', 'starmap_n']

import atexit
import enum
import os
import signal
import sys
import warnings
import weakref
from collections.abc import Callable, Iterable, Iterator, Mapping, Sized
from concurrent.futures import Executor, Future
from concurrent.futures import TimeoutError as _TimeoutError
from contextlib import ExitStack, contextmanager
from cProfile import Profile
from functools import partial
from itertools import chain, islice, starmap
from multiprocessing.managers import BaseManager
from operator import methodcaller
from pstats import Stats
from queue import Empty, SimpleQueue
from threading import Lock
from time import perf_counter, sleep
from typing import Protocol, TypeVar, cast

import loky

try:
    import psutil
except ImportError:
    psutil = None

from ._more import chunked, ilen
from ._reduction import move_to_shmem, reducers
from ._thread_quota import ThreadQuota

_T = TypeVar('_T')
_T_ctr = TypeVar('_T_ctr', contravariant=True)
_K = TypeVar('_K')
_F = TypeVar('_F', bound=Future)

_NUM_CPUS = os.cpu_count() or 0
_NUM_CPUS = min(_NUM_CPUS, int(os.getenv('GLOW_CPUS', _NUM_CPUS)))
_IDLE_WORKER_TIMEOUT = 10
_GRANULAR_SCHEDULING = False  # TODO: investigate whether this improves load


class _Empty(enum.Enum):
    token = 0


_empty = _Empty.token

# ------------------- some useful interfaces and functions -------------------


class _Queue(Protocol[_T]):
    def get(self, block: bool = ..., timeout: float | None = ...) -> _T:
        ...

    def put(self, item: _T) -> None:
        ...


class _Event(Protocol):
    def is_set(self) -> bool:
        ...

    def set(self) -> None:
        ...


def _get_cpu_count_limits(upper_bound: int = sys.maxsize,
                          mp: bool = False) -> Iterator[int]:
    yield upper_bound
    yield os.cpu_count() or 1

    # Windows platform lacks memory overcommit, so it's sensitive to VMS growth
    if not mp or sys.platform != 'win32' or 'torch' not in sys.modules:
        return

    import torch
    if torch.version.cuda and torch.version.cuda >= '11.7.0':
        # It's expected that torch will fix .nv_fatb readonly flag in its DLLs
        # See https://stackoverflow.com/a/69489193/9868257
        return

    if psutil is None:
        warnings.warn(
            'Max process count may be calculated incorrectly, '
            'leading to application crash or even BSOD. '
            'Install psutil to avoid that',
            stacklevel=3)
        return

    # Overcommit on Windows is forbidden, thus VMS planning is necessary
    vms: int = psutil.Process().memory_info().vms
    free_vms: int = psutil.virtual_memory().free + psutil.swap_memory().free
    yield free_vms // vms


def max_cpu_count(upper_bound: int = sys.maxsize, mp: bool = False) -> int:
    return min(_get_cpu_count_limits(upper_bound, mp))


_PATIENCE = 0.01


def _retry_call(fn: Callable[..., _T], *exc: type[BaseException]) -> _T:
    # See issues
    # https://bugs.python.org/issue29971
    # https://github.com/python/cpython/issues/74157
    # https://github.com/dask/dask/pull/2144#issuecomment-290556996
    # https://github.com/dask/dask/pull/2144/files
    while True:
        try:
            return fn(timeout=_PATIENCE)
        except exc:
            sleep(0)  # Force switch to another thread to proceed


if sys.platform == 'win32':

    def _result(f: Future[_T], /) -> _T:
        return _retry_call(f.result, _TimeoutError)
else:
    _result = Future.result


def _result_or_cancel(f: Future[_T]) -> _T:
    try:
        try:
            return _result(f)
        finally:
            f.cancel()
    finally:
        del f


def _q_get_fn(q: _Queue[_T]) -> Callable[[], _T]:
    if sys.platform != 'win32':
        return q.get
    return partial(_retry_call, q.get, Empty)


# ---------------------------- pool initialization ----------------------------


def _mp_profile():
    """Multiprocessed profiler"""
    prof = Profile()
    prof.enable()

    def _finalize(lines=50):
        prof.disable()
        with open(f'prof-{os.getpid()}.txt', 'w') as fp:
            Stats(prof, stream=fp).sort_stats('cumulative').print_stats(lines)

    atexit.register(_finalize)


def _initializer():
    # `signal.signal` suppresses KeyboardInterrupt in child processes
    signal.signal(signal.SIGINT, signal.SIG_IGN)
    if os.environ.get('_GLOW_MP_PROFILE'):
        _mp_profile()


@contextmanager
def _get_executor(max_workers: int, mp: bool) -> Iterator[Executor]:
    if mp:
        processes: loky.ProcessPoolExecutor = loky.get_reusable_executor(
            max_workers,
            'loky_init_main',
            _IDLE_WORKER_TIMEOUT,
            job_reducers=reducers,
            result_reducers=reducers,
            initializer=_initializer,
        )
        # In generator 'finally' is not reliable enough, use atexit
        hook = atexit.register(processes.shutdown, kill_workers=True)
        yield processes
        atexit.unregister(hook)
    else:
        threads = ThreadQuota(max_workers)
        try:
            yield threads
        finally:
            is_success = sys.exc_info()[0] is None
            threads.shutdown(wait=is_success, cancel_futures=True)


def _get_manager(executor: Executor):
    if isinstance(executor, loky.ProcessPoolExecutor):
        return executor._context.Manager()

    else:  # noqa: RET505
        from multiprocessing.dummy import Manager
        return Manager()


# -------- bufferize iterable by offloading to another thread/process --------


def _consume(iterable: Iterable[_T], q: _Queue[_T | _Empty], ev: _Event):
    try:
        for item in iterable:
            if ev.is_set():
                break
            q.put(item)
    finally:
        q.put(_empty)  # Signal to stop iteration
        q.put(_empty)  # Match last q.get


class buffered(Iterator[_T]):  # noqa: N801
    """
    Iterates over `iterable` in background thread with at most `latency`
    items ahead from caller
    """
    __slots__ = ('_get', '_task', 'close', '__weakref__')

    def __init__(self,
                 iterable: Iterable[_T],
                 /,
                 *,
                 latency: int = 2,
                 mp: bool | Executor = False):
        s = ExitStack()
        if isinstance(mp, Executor):
            executor = mp
        else:
            executor = s.enter_context(_get_executor(1, mp))

        mgr = _get_manager(executor)
        if isinstance(mgr, BaseManager):
            s.enter_context(mgr)

        ev: _Event = mgr.Event()
        q: _Queue[_T | _Empty] = mgr.Queue(latency)
        self._task = executor.submit(_consume, iterable, q, ev)  # type: ignore
        self._get = q_get = _q_get_fn(q)

        # If main killed, wakes up consume to check ev
        # Otherwise collects 2nd _empty from q.
        # Called 2nd
        s.callback(q_get)

        # If main killed, signals consume to stop.
        # If consume is already stopped (on error or not), does nothing.
        # Called 1st
        s.callback(ev.set)

        self.close = weakref.finalize(self, s.close)

    def __iter__(self) -> Iterator[_T]:
        return self

    def __next__(self) -> _T:
        if self.close.alive:
            if (item := self._get()) is not _empty:
                return item

            self.close()
            _result(self._task)  # Throws exception if worker killed itself

        raise StopIteration


# ---------------------------- automatic batching ----------------------------


class _AutoSize:
    MIN_DURATION = 0.2
    MAX_DURATION = 2.0
    size: int = 1
    duration: float = 0.0

    def __init__(self) -> None:
        self.lock = Lock()

    def suggest(self) -> int:
        with self.lock:
            if 0 < self.duration < self.MIN_DURATION:
                self.size *= 2
                self.duration = 0.0

            elif self.duration > self.MAX_DURATION:
                size = int(2 * self.size * self.MIN_DURATION / self.duration)
                size = max(size, 1)
                if self.size != size:
                    self.duration = 0.0
                    self.size = size

            return self.size

    def update(self, start_time: float, fut: Future[Sized]):
        # Compute as soon as future became done, discard later if not needed
        duration = perf_counter() - start_time

        try:
            # Cannot time out, as future is always completed at this point.
            # Though it's unknown whether it succeeded or not, so use EAFP
            r = fut.result()
        except BaseException:  # noqa: BLE001
            return

        with self.lock:
            if len(r) != self.size:
                return
            self.duration = ((0.8 * self.duration + 0.2 * duration)
                             if self.duration > 0 else duration)


# ---------------------- map iterable through function ----------------------


def _schedule(make_future: Callable[..., _F], args_zip: Iterable[Iterable],
              chunksize: int) -> Iterator[_F]:
    return starmap(make_future, chunked(args_zip, chunksize))


def _schedule_auto(
    make_future: Callable[..., _F],
    args_zip: Iterator[Iterable],
    max_workers: int,
) -> Iterator[_F]:
    # For the whole wave make futures with the same job size
    size = _AutoSize()
    while tuples := [*islice(args_zip, size.suggest() * max_workers)]:
        chunksize = len(tuples) // max_workers or 1
        for f in starmap(make_future, chunked(tuples, chunksize)):
            f.add_done_callback(partial(size.update, perf_counter()))
            yield f


def _schedule_auto_v2(make_future: Callable[..., _F],
                      args_zip: Iterator[Iterable]) -> Iterator[_F]:
    # Vary job size from future to future
    size = _AutoSize()
    while args := [*islice(args_zip, size.suggest())]:
        f = make_future(*args)
        f.add_done_callback(partial(size.update, perf_counter()))
        yield f


def _get_unwrap_iter(s: ExitStack, qsize: int,
                     get_done_f: Callable[[], Future[_T]],
                     fs_scheduler: Iterator) -> Iterator[_T]:
    with s:
        if not qsize:  # No tasks to do
            return

        # Unwrap 1st / schedule `N-qsize` / unwrap `qsize-1`
        for _ in chain([None], fs_scheduler, range(qsize - 1)):

            # Retrieve done task, exactly `N` calls
            yield _result_or_cancel(get_done_f())


def _unwrap(s: ExitStack, fs: Iterable[Future[_T]], qsize: int | None,
            order: bool) -> Iterator[_T]:
    q = SimpleQueue[Future[_T]]()

    # If `order`, then `q` has "PENDING"/"RUNNING"/"DONE" tasks,
    # otherwise it only has "DONE" tasks.
    # FIXME: order=False -> random freezes (in q.get -> Empty)
    q_put = cast(Callable[[Future[_T]], None],
                 q.put if order else methodcaller('add_done_callback', q.put))

    # On each `next()` schedules new task
    fs_scheduler = map(q_put, fs)
    try:
        # Fetch up to `qsize` tasks to pre-fill `q`
        qsize = ilen(islice(fs_scheduler, qsize))

    except BaseException:
        # Unwind stack here on an error
        s.close()
        raise

    else:
        return _get_unwrap_iter(s, qsize, _q_get_fn(q), fs_scheduler)


def _batch_invoke(func: Callable[..., _T], *items: tuple) -> list[_T]:
    return [*starmap(func, items)]


def starmap_n(func: Callable[..., _T],
              iterable: Iterable[Iterable],
              /,
              *,
              max_workers: int | None = None,
              prefetch: int | None = 2,
              mp: bool = False,
              chunksize: int | None = None,
              order: bool = True) -> Iterator[_T]:
    """
    Equivalent to itertools.starmap(fn, iterable).

    Return an iterator whose values are returned from the function evaluated
    with an argument tuple taken from the given sequence.

    Options:

    - workers - Count of workers, by default all hardware threads are occupied.
    - prefetch - Extra count of scheduled jobs, if not set equals to infinity.
    - mp - Whether use processes or threads.
    - chunksize - The size of the chunks the iterable will be broken into
      before being passed to a processes. Estimated automatically.
      Ignored when threads are used.
    - order - Whether keep results order, or ignore it to increase performance.

    Unlike multiprocessing.Pool or concurrent.futures.Executor this one:

    - never deadlocks on any exception or Ctrl-C interruption.
    - accepts infinite iterables due to lazy task creation (option prefetch).
    - has single interface for both threads and processes.
    - TODO: serializes array-like data using out-of-band Pickle 5 buffers.
    - before first `__next__` call it submits at most `prefetch` jobs
      to warmup pool of workers.

    Notes:

    - To reduce latency set order to False, order of results will be arbitrary.
    - To increase CPU usage increase prefetch or set it to None.
    - In terms of CPU usage there's no difference between
      prefetch=None and order=False, so choose wisely.
    - Setting order to False makes no use of prefetch more than 0.

    TODO: replace `order=True` with `heap=False`
    """
    if max_workers is None:
        max_workers = max_cpu_count(_NUM_CPUS, mp)

    if not max_workers or not _NUM_CPUS:
        return starmap(func, iterable)  # Fallback to single thread

    if mp and chunksize is None and prefetch is None:
        raise ValueError('With multiprocessing either chunksize or prefetch '
                         'must be not None')

    if prefetch is not None:
        prefetch = max(prefetch + max_workers, 1)

    it = iter(iterable)
    s = ExitStack()
    submit = s.enter_context(_get_executor(max_workers, mp)).submit

    if mp:
        func = move_to_shmem(func)
    else:
        chunksize = chunksize or 1

    if chunksize == 1:
        submit_one = cast(
            Callable[..., Future[_T]],
            partial(submit, func),
        )
        return _unwrap(s, starmap(submit_one, it), prefetch, order)

    submit_many = cast(
        Callable[..., Future[list[_T]]],
        partial(submit, _batch_invoke, func),
    )
    if chunksize is not None:
        # Fixed chunksize
        fs = _schedule(submit_many, it, chunksize)
    elif not _GRANULAR_SCHEDULING:
        # Dynamic chunksize scaling, submit tasks in waves
        fs = _schedule_auto(submit_many, it, max_workers)
    else:
        # Dynamic chunksize scaling
        fs = _schedule_auto_v2(submit_many, it)

    chunks = _unwrap(s, fs, prefetch, order)
    return chain.from_iterable(chunks)


def map_n(func: Callable[..., _T],
          /,
          *iterables: Iterable,
          max_workers: int | None = None,
          prefetch: int | None = 2,
          mp: bool = False,
          chunksize: int | None = None,
          order: bool = True) -> Iterator[_T]:
    """
    Returns iterator equivalent to map(func, *iterables).

    Make an iterator that computes the function using arguments from
    each of the iterables. Stops when the shortest iterable is exhausted.

    For extra options, see starmap_n, which is used under hood.
    """
    return starmap_n(
        func,
        zip(*iterables),
        max_workers=max_workers,
        prefetch=prefetch,
        mp=mp,
        chunksize=chunksize,
        order=order)


def map_n_dict(func: Callable[[_T_ctr], _T],
               obj: Mapping[_K, _T_ctr],
               /,
               *,
               max_workers: int | None = None,
               prefetch: int | None = 2,
               mp: bool = False,
               chunksize: int | None = None) -> dict[_K, _T]:
    """
    Apply `func` to each value in a mapping in parallel way.
    For extra options, see starmap_n, which is used under hood.
    """
    iter_values = map_n(
        func,
        obj.values(),
        max_workers=max_workers,
        prefetch=prefetch,
        mp=mp,
        chunksize=chunksize)
    return dict(zip(obj.keys(), iter_values))



from __future__ import annotations

__all__ = ['Reusable']

import asyncio
import sys
import weakref
from collections.abc import Callable
from dataclasses import dataclass, field
from functools import partial
from threading import Thread
from typing import Generic, TypeVar

from .concurrency import call_once

_T = TypeVar('_T')
_Make = Callable[[], _T]
_Callback = Callable[[_T], object]


@call_once
def make_loop() -> asyncio.AbstractEventLoop:
    loop = asyncio.new_event_loop()
    Thread(target=loop.run_forever, daemon=True).start()
    return loop


async def _await(fn: _Make[_T]) -> _T:
    return fn()


def _trampoline(callback: _Callback[_T], ref: weakref.ref[_T]) -> None:
    if (obj := ref()) is not None:
        callback(obj)


@dataclass
class Reusable(Generic[_T]):
    make: _Make[_T]
    delay: float
    finalize: _Callback[_T] | None = None

    _loop: asyncio.AbstractEventLoop = field(default_factory=make_loop)
    _deleter: asyncio.TimerHandle | None = None
    _box: list[_T] = field(default_factory=list)

    if sys.version_info >= (3, 10):
        _lock: asyncio.Lock = field(default_factory=asyncio.Lock)
    else:
        _lock: asyncio.Lock = field(init=False)

        def __post_init__(self):
            self._lock = asyncio.run_coroutine_threadsafe(
                _await(asyncio.Lock),
                self._loop,
            ).result()
            assert self._lock._loop is self._loop

    def __call__(self) -> _T:
        """Returns inner object, or recreates it"""
        fut = asyncio.run_coroutine_threadsafe(self._get(), self._loop)
        return fut.result()

    async def _get(self) -> _T:
        async with self._lock:
            # reschedule finalizer
            if self._deleter is not None:
                self._deleter.cancel()
            self._deleter = self._loop.call_later(self.delay, self._box.clear)

            # recreate object, if nessessary
            if not self._box:
                obj = await asyncio.to_thread(self.make)
                if self.finalize is not None:
                    weakref.ref(obj, partial(_trampoline, self.finalize))
                self._box.append(obj)

            return self._box[0]



from __future__ import annotations

__all__ = [
    'auto_ddp', 'auto_model', 'barrier', 'get_ddp_info', 'reduce_if_needed'
]

import pickle
from collections.abc import Callable
from functools import partial, update_wrapper
from pathlib import Path
from typing import Any, NamedTuple, Protocol, TypeVar, cast

import torch
import torch.cuda
import torch.distributed as dist
import torch.multiprocessing as tmp
from torch import nn

# -------------------------------- primitives --------------------------------


class _DdpInfo(NamedTuple):
    world: int
    rank: int


def get_ddp_info() -> _DdpInfo | None:
    if not dist.is_initialized():
        return None
    return _DdpInfo(dist.get_rank(), dist.get_world_size())


def barrier(rank: int | None = None) -> None:
    """Synchronize all processes"""
    if (info := get_ddp_info()) and (rank is None or rank == info.rank):
        dist.barrier()


def reduce_if_needed(*tensors: torch.Tensor,
                     mean: bool = False) -> tuple[torch.Tensor, ...]:
    """Reduce tensors across all machines"""
    if (ddp := get_ddp_info()) and ddp.world > 1:
        tensors = *(t.clone() for t in tensors),
        dist.all_reduce_multigpu(tensors)
        if mean:
            tensors = *(t / ddp.world for t in tensors),
    return tensors


# --------------------------------- wrappers ---------------------------------


def auto_model(net: nn.Module, sync_bn: bool = True) -> nn.Module:
    if (ddp := get_ddp_info()) and ddp.world > 1:
        torch.cuda.set_device(ddp.rank)

        net.to(ddp.rank)
        if sync_bn:
            net = nn.SyncBatchNorm.convert_sync_batchnorm(net)
        return nn.parallel.DistributedDataParallel(net, device_ids=[ddp.rank])

    net.cuda()
    return (nn.parallel.DataParallel(net)
            if torch.cuda.device_count() > 1 else net)


class _TrainFn(Protocol):
    def __call__(self, __net: nn.Module, *args, **kwargs) -> Any:
        ...


_F = TypeVar('_F', bound=Callable)
_TrainFnType = TypeVar('_TrainFnType', bound=_TrainFn)


class _AutoDdp:
    def __init__(self, train_fn: _TrainFn, net: nn.Module, *args, **kwargs):
        self.train_fn = train_fn
        self.net = net
        self.args = args
        self.kwargs = kwargs
        self.ngpus = torch.cuda.device_count()

        if self.ngpus == 1:
            self._worker(None)
            return

        # ! Not tested
        # * Actually, here we can use loky.ProcessPoolExecutor, like this:
        # from . import map_n
        # ngpus = self.ngpus
        # jobs = map_n(self._worker, range(ngpus), max_workers=ngpus, mp=True)
        # list(jobs)
        # * Left as safe measure
        tmp.spawn(self._worker, nprocs=self.ngpus)

    def _worker(self, rank: int | None) -> None:
        if rank is None:
            return self.train_fn(self.net, *self.args, **self.kwargs)

        dist.init_process_group(
            backend='nccl', rank=rank, world_size=self.ngpus)
        try:
            self.train_fn(auto_model(self.net), *self.args, **self.kwargs)
        finally:
            dist.destroy_process_group()


def auto_ddp(train_fn: _TrainFnType) -> _TrainFnType:
    return cast(_TrainFnType,
                update_wrapper(partial(_AutoDdp, train_fn), train_fn))


def once_per_world(fn: _F) -> _F:
    """Call function only in rank=0 process, and share result for others"""
    def wrapper(*args, **kwargs):
        ddp = get_ddp_info()
        if not ddp or ddp.world == 1:
            # Master process, so no neighbors to share results with
            return fn(*args, **kwargs)

        # Generate random fname and share it among whole world
        idx = torch.empty((), dtype=torch.int64).random_()
        if ddp.rank == 0:
            dist.broadcast(idx, 0)

        tmp = Path(f'/tmp/_ddp_share_{idx.item():x}.pkl')
        result = None

        if ddp.rank == 0:  # 0th child
            result = fn(*args, **kwargs)
            with tmp.open('wb') as fp:
                pickle.dump(result, fp)

        barrier()

        if ddp.rank > 0:  # Gather results from 0th child
            with tmp.open('rb') as fp:
                result = pickle.load(fp)

        barrier()

        if ddp.rank == 0:  # 0th child
            tmp.unlink()

        return result

    return cast(_F, update_wrapper(wrapper, fn))



from __future__ import annotations

__all__ = ['get_grads']

import warnings
from collections import defaultdict
from collections.abc import Iterable
from contextlib import contextmanager

import torch
from torch import nn, optim

try:
    from .driver import get_gpu_capability
except ImportError:
    get_gpu_capability = None  # type: ignore[assignment]

_MIN_SCALE = 2.0 ** -16
_MAX_SCALE = 2.0 ** +16
_PATIENCE = 2000

_MIN_FP16_CUDA_CAPABILITY = (7, 0)


def _fp16_is_definitely_not_available() -> bool:
    """
    Checks if all GPUs are not CUDA 7.x+ capable (to enable FP16).
    Doesn't trigger CUDA init.

    Possible results:
    - CUDA capability <7.0 -> FP32 mode
    - CUDA capability >=7.0,<8.0 -> FP32 mode, FP16 with scaler
    - CUDA capability >=8.0 -> FP32/TF32/BF16 as is, FP16 with scaler
    """
    if not torch.cuda.is_available():
        # No CUDA, so definitely no FP16
        return True

    if not torch.cuda.is_initialized():
        # Using Torch to check CUDA capability here triggers CUDA init,
        # (what is heavy if done from Torch side due to lots of kernels)
        # but we have no NVML alternative,
        # so consider that *maybe* some of GPUs can FP16
        if get_gpu_capability is None:
            return False

        # Rely on NVML driver
        caps = get_gpu_capability()

    else:
        # CUDA is initialized already, no need to be careful
        device_ids = range(torch.cuda.device_count())
        caps = [torch.cuda.get_device_capability(dev) for dev in device_ids]

    return all(cap < _MIN_FP16_CUDA_CAPABILITY for cap in caps)


# ------------------------- basic context, no scaler -------------------------


class Grads:
    _num_backwards: int = 0

    def __init__(self,
                 opt: optim.Optimizer,
                 sched: optim.lr_scheduler._LRScheduler | None = None):
        self._opt = opt
        self._sched = sched

    def zero_grad(self) -> None:
        self._opt.zero_grad()

    def backward(self, tensor: torch.Tensor):
        tensor.backward()
        self._num_backwards += 1

    def unscale(self) -> None:
        return

    def step(self) -> None:
        if self._num_backwards:
            self._opt.step()
            if self._sched:
                self._sched.step()
        self._num_backwards = 0

    def __enter__(self) -> None:
        self.zero_grad()

    def __exit__(self, type_, *_) -> None:
        if type_ is None:
            self.step()

    def state_dict(self) -> dict:
        s = {
            'num_backwards': self._num_backwards,
            'optimizer': self._opt.state_dict(),
        }
        if self._sched is not None:
            s['scheduler'] = self._sched.state_dict()
        return s

    def load_state_dict(self, state: dict):
        self._num_backwards = state['num_backwards']
        self._opt.load_state_dict(state['optimizer'])
        if self._sched is not None:
            self._sched.load_state_dict(state['scheduler'])


# --------------------------- default grad scaler ----------------------------
_PRIVATE = True
_NAN = float('nan')  # nan != nan


class _ScalingGrads(Grads):
    """
    Grads around torch's GradScaler.

    - Accumulates scaled grads in .grad fields
    - Unscaling and scale update done once per optim.step
    - If any grad fraction has inf/nan, then optim.step is skipped,
      thus making whole sequence of N fwd-bwd calls (N = grad steps)
      gone to nowhere.
    """
    def __init__(self,
                 opt: optim.Optimizer,
                 sched: optim.lr_scheduler._LRScheduler | None = None,
                 scale: float = _MAX_SCALE):
        super().__init__(opt, sched)
        self._scaler = torch.cuda.amp.grad_scaler.GradScaler(scale)

        if _PRIVATE:
            self._steps = getattr(self._opt, '_step_count', _NAN)
        else:
            self._scale = self._scaler.get_scale()

    def backward(self, tensor: torch.Tensor) -> None:
        self._scaler.scale(tensor).backward()

    def unscale_(self) -> None:
        self._scaler.unscale_(self._opt)

    def step(self) -> None:
        self._scaler.step(self._opt)
        self._scaler.update()

        if not self._sched:
            return

        if _PRIVATE:
            steps, self._steps = self._steps, getattr(self._opt, '_step_count',
                                                      _NAN)
            if steps == self._steps:  # No step done
                return
        else:
            # ! CPU-GPU sync
            old, self._scale = self._scale, self._scaler.get_scale()
            if old > self._scale:  # Scale was decreased due to overflow
                return

        self._sched.step()

    def state_dict(self) -> dict:
        return super().state_dict() | {'scaler': self._scaler.state_dict()}

    def load_state_dict(self, state: dict):
        super().load_state_dict(state)
        self._scaler.load_state_dict(state['scaler'])


# ------------------------- grad scaler with retries -------------------------


@contextmanager
@torch.no_grad()
def accumulating(params: Iterable[torch.Tensor], done: int):
    def _borrow(p: torch.Tensor) -> torch.Tensor | None:
        # Save old state
        grad, p.grad = p.grad, None
        return None if grad is None else grad.detach_()

    def _update(p: torch.Tensor, grad: torch.Tensor | None):
        # In case the fail happened, or new grad is missing, use old grad
        if inf[0] or p.grad is None:
            p.grad = grad

        # If both grads exist, do running mean update
        elif grad is not None:
            # mu' = (new + mu * n) / (1 + n)
            #     = lerp(new, mu, n / (n + 1))
            p.grad.lerp_(grad, done / (1 + done))

    stash = {p: _borrow(p) for p in params}
    inf = [False]
    try:
        with torch.enable_grad():
            yield inf
    finally:
        for p, grad in stash.items():
            _update(p, grad)


class _GenericScalingGrads(Grads):
    """
    Grads around custom GradScaler (inlined and optimized).

    Differences from torch.GradScaler:

    - Stores scaled grads from single bwd in .grad fields
    - Unscales grads and updates scale
    - If any .grad got inf/nan, does backward() again to get correct grads,
      (configurable), or drops bad grads.
    - Mixes new grads with the grads from the previous calls
    - After N such fwd-bwd calls does step if all grads in sequence were ok.

    Thus, this scaler always tries to re-compute failed backward(),
    reducing chance of skipping optimizer step.
    But does gradient unscaling more frequiently, and maybe this can add some
    CPU-GPU sync penalty.

    This scaler takes care of consecutive bwd calls, averaging gradients
    over time.
    """
    def __init__(self,
                 opt: optim.Optimizer,
                 sched: optim.lr_scheduler._LRScheduler | None = None,
                 max_retries: int = 1,
                 scale: float = _MAX_SCALE,
                 min_scale: float | None = _MIN_SCALE):
        self._max_retries = max_retries
        self._min_scale = min_scale

        self._growth_tracker = torch.zeros(1).int()
        self._scale = torch.empty(1).fill_(scale)

        self._params: set[nn.Parameter] = {
            p for param_group in opt.param_groups
            for p in param_group['params'] if p.requires_grad
        }
        super().__init__(opt, sched)

    def _unscale_grads_(self) -> tuple[torch.Tensor, ...]:
        devs = set[torch.device]()
        grad_groups: dict[tuple[torch.device, torch.dtype],
                          list[torch.Tensor]] = defaultdict(list)
        for p in self._params:
            if p.grad is None:
                continue
            if p.grad.dtype == torch.float16:
                raise ValueError('Attempting to unscale FP16 gradients.')
            devs.add(p.device)
            grad_groups[p.device, p.dtype].append(p.grad)

        # Compute in FP64 for more precision
        inv_scale = self._scale.double().reciprocal().float()

        # Inf flags
        infs = {dev: torch.zeros(1, device=dev) for dev in devs}

        # Unscale grads and update infs
        for (dev, _), grads in grad_groups.items():
            torch._amp_foreach_non_finite_check_and_unscale_(
                grads, infs[dev], inv_scale.to(dev, non_blocking=True))
        return *infs.values(),

    def _update_(self, infs: Iterable[torch.Tensor]) -> None:
        # Collect infs from all devices
        found_inf, *rest = (
            inf.to(self._scale.device, non_blocking=True) for inf in infs)
        for inf in rest:
            found_inf += inf

        torch._amp_update_scale_(self._scale, self._growth_tracker, found_inf,
                                 2.0, 0.5, _PATIENCE)

    def backward(self, tensor: torch.Tensor) -> None:
        self._growth_tracker = self._growth_tracker.to(
            tensor.device, non_blocking=True)
        self._scale = self._scale.to(tensor.device, non_blocking=True)

        # Borrow grads, and zero source
        with accumulating(self._params, self._num_backwards) as inf:
            for remaining in reversed(range(1 + self._max_retries)):
                # Collect grads from backward
                (tensor * self._scale).backward(retain_graph=bool(remaining))

                # Check for inf/nan
                infs = self._unscale_grads_()
                self._update_(infs)
                inf[0] = any(v.item() for v in infs)  # ! cpu-gpu sync

                if not inf[0]:  # Success
                    self._num_backwards += 1
                    return

                # Second round (or complete failure), zero grads to fill again
                self.zero_grad()
                print('Got inf in grads,',
                      'try another scale' if remaining else 'skip step')

        # ! Optional cpu-gpu sync
        if self._min_scale is None or self._scale.item() > self._min_scale:
            return
        raise OverflowError(f'Scale underflow {self._min_scale}')

    def state_dict(self) -> dict:
        return super().state_dict() | {
            'max_retries': self._max_retries,
            'min_scale': self._min_scale,
            'scale': self._scale.item(),
            'growth_tracker': self._growth_tracker.item()
        }

    def load_state_dict(self, state: dict):
        super().load_state_dict(state)
        self._max_retries = state['max_retries']
        self._min_scale = state['min_scale']
        self._growth_tracker.fill_(state['growth_tracker'])
        self._scale.fill_(state['scale'])
        self._params = {
            p for param_group in self._opt.param_groups
            for p in param_group['params'] if p.requires_grad
        }


def get_grads(opt: optim.Optimizer,
              sched: optim.lr_scheduler._LRScheduler | None = None,
              dtype: torch.dtype | None = None,
              max_retries: int = 1) -> Grads:
    """Get gradients context with specified precision

    Parameters:
    - dtype - dtype used for computation. If float16 used, uses GradScaler
      to prevent nan/inf during backward passes.
      By default determines whether to enable fp16
      depending on CUDA capability of available CUDA devices.
    - max_retries - number of additional attempts to use another scale
      for the same loss when backward() results to NaN/Inf. Only for fp16 mode.

    Example:
    ```
    # Creates model and optimizer in default precision
    model = Net()
    optimizer = optim.SGD(model.parameters(), ...)

    # Enable amp
    grads = gnn.get_grads(optimizer, dtype=torch.half)
    autocast = torch.autocast(device.type, torch.half)

    with grads:
        for input, target in batches:
            with autocast:
                output = model(input)
                loss = loss_fn(output, target)

            grads.backward(loss)

        grads.unscale_()  # No-op for some Grads subclasses
        torch.nn.utils.clip_grad_norm_(opt.parameters(), MAX_NORM)
    ```
    Notes:
    - never call model.half() as autocast doesn't need FP16 weights/buffers
      unless inference mode is needed.
    - for inference mode, do weight-norm fusion before using FP16
    - by default torch (1.6-1.12) doesn't promote batchnorm/instancenorm to
      FP32 (unlike layernorm/groupnorm), thus it may cause stability issues.
      Unless you'll find a way to register batchnorm as op promoting to FP32
      in autocast mode.
      https://pytorch.org/docs/stable/amp.html#ops-that-can-autocast-to-float32
    """
    if _fp16_is_definitely_not_available():  # None can FP16, disable it anyway
        if dtype != torch.float:
            warnings.warn('CUDA cannot FP16, fallback to FP32', stacklevel=2)
        dtype = torch.float
    elif dtype is None:  # Some GPU's can do FP16, automatic choice, enable
        dtype = torch.half

    if dtype != torch.half:
        return Grads(opt, sched)
    if not max_retries:
        return _ScalingGrads(opt, sched)
    return _GenericScalingGrads(opt, sched, max_retries)



from __future__ import annotations

__all__ = ['move_to_shmem']

import copyreg
import io
import logging
import mmap
import os
import pickle
import sys
import tempfile
import weakref
from collections.abc import Callable
from multiprocessing.shared_memory import SharedMemory
from pathlib import Path
from typing import Generic, TypeVar

import loky

from ._import_hook import when_imported

_T = TypeVar('_T')
_F = TypeVar('_F', bound=Callable)

_SYSTEM_SHM_MIN_SIZE = int(2e9)
_SYSTEM_SHM = Path('/dev/shm')
_SYSTEM_TEMP = Path(tempfile.gettempdir())

reducers: dict[type, Callable] = {}
loky.set_loky_pickler('pickle')

logger = logging.getLogger(__name__)
# logger.setLevel(logging.DEBUG)
# logger.addHandler(logging.StreamHandler())


def _get_shm_dir() -> Path:
    if sys.platform != 'win32':
        try:
            stats = os.statvfs(_SYSTEM_SHM)
            if stats.f_bsize * stats.f_bavail > _SYSTEM_SHM_MIN_SIZE:
                return _SYSTEM_SHM
        except OSError:
            pass
    return _SYSTEM_TEMP


class _Proxy(Generic[_F]):
    call: _F

    def __call__(self, *args):
        return self.call(*args)


class _NullProxy(_Proxy[_F]):
    __slots__ = ('call', )

    def __init__(self, call: _F):
        self.call = call


class _Mmap:
    __slots__ = ('size', 'tag', 'buf', '__weakref__')
    _shm_root = _get_shm_dir()

    @classmethod
    def from_bytes(cls, data: memoryview, tag: str) -> _Mmap:
        mv = cls(data.nbytes, f'shm-{os.getpid()}-{tag}', create=True)
        mv.buf[:] = data
        return mv

    def __init__(self, size, tag, create=False):
        self.size = size
        self.tag = tag
        if create:
            flags = os.O_CREAT | os.O_EXCL | os.O_RDWR
            access = mmap.ACCESS_WRITE
        else:
            flags, access = os.O_RDONLY, mmap.ACCESS_READ

        if sys.platform == 'win32':
            self.buf = mmap.mmap(-1, self.size, self.tag, access=access)
        else:
            name = self._shm_root / f'psm_{self.tag}'
            fd = os.open(name, flags, 0o600)
            # resource_tracker.register(name.as_posix(), 'file')
            if create:
                os.ftruncate(fd, self.size)

            self.buf = mmap.mmap(fd, self.size, access=access)
            if create:
                weakref.finalize(self.buf, os.unlink, name)
            # weakref.finalize(self.buf, resource_tracker.maybe_unlink,
            #                  name.as_posix(), 'file')
            weakref.finalize(self.buf, os.close, fd)

    def __reduce__(self):
        return type(self), (self.size, self.tag)

    def __sizeof__(self):
        return self.size


@when_imported('torch')
def _torch_hook(torch):
    reducers.update({
        torch.Tensor: torch.multiprocessing.reductions.reduce_tensor,
        **{
            t: torch.multiprocessing.reductions.reduce_storage
            for t in torch.storage._StorageBase.__subclasses__()
        },
    })


def _dumps(
    obj: object,
    callback: Callable[[pickle.PickleBuffer], object] | None = None,
) -> bytes:
    fp = io.BytesIO()
    p = pickle.Pickler(fp, -1, buffer_callback=callback)
    p.dispatch_table = copyreg.dispatch_table | reducers
    p.dump(obj)
    return fp.getvalue()


def _mmap_reconstruct(data: bytes, memos: list[_Mmap]):
    buffers = [m.buf for m in memos]
    return pickle.loads(data, buffers=buffers)


class _MmapProxy(_Proxy[_F]):
    __slots__ = ('uid', 'call', 'data', 'memos')

    def __init__(self, call: _F) -> None:
        buffers: list[pickle.PickleBuffer] = []
        self.uid = id(call)
        self.call = call
        self.data = _dumps(call, callback=buffers.append)
        self.memos = []
        for i, buf in enumerate(buffers):
            with buf.raw() as m:
                self.memos += [_Mmap.from_bytes(m, f'{self.uid:x}-{i}')]

    def __reduce__(self) -> tuple:
        return _mmap_reconstruct, (self.data, self.memos)


def _shn_reconstruct(data: bytes, memos: list[tuple[SharedMemory, int]]):
    buffers = [sm.buf[:size] for sm, size in memos]
    return pickle.loads(data, buffers=buffers)


class _ShmemProxy(_Proxy[_F]):
    __slots__ = ('call', 'data', 'memos')

    def __init__(self, call: _F) -> None:
        buffers: list[pickle.PickleBuffer] = []
        self.call = call
        self.data = _dumps(call, callback=buffers.append)
        self.memos = []
        for buf in buffers:
            with buf.raw() as m:
                sm = SharedMemory(create=True, size=m.nbytes)
                sm.buf[:] = m
                self.memos.append((sm, sm.size))

    def __reduce__(self) -> tuple:
        return _shn_reconstruct, (self.data, self.memos)


def move_to_shmem(fn: Callable[..., _T]) -> Callable[..., _T]:
    return _NullProxy(fn)
    return _ShmemProxy(fn)



from __future__ import annotations

__all__ = [
    'as_iter', 'chunked', 'eat', 'ichunked', 'ilen', 'roundrobin', 'windowed'
]

import threading
from collections import deque
from collections.abc import Iterable, Iterator, Mapping, Sequence, Sized
from functools import partial
from itertools import chain, count, cycle, islice, repeat
from typing import Protocol, TypeVar, overload

_T = TypeVar('_T')
_T_co = TypeVar('_T_co', covariant=True)


class SupportsSlice(Sized, Protocol[_T_co]):
    def __getitem__(self, __s: slice) -> _T_co:
        ...


# ----------------------------------------------------------------------------


def as_iter(obj: Iterable[_T] | _T, limit: int | None = None) -> Iterator[_T]:
    """Make iterator with at most `limit` items"""
    if isinstance(obj, Iterable):
        return islice(obj, limit)
    return repeat(obj) if limit is None else repeat(obj, limit)


# ----------------------------------------------------------------------------


def _dispatch(fallback_fn, fn, it, *args):
    if (not isinstance(it, Sized) or not hasattr(it, '__getitem__')
            or isinstance(it, Mapping)):
        return fallback_fn(it, *args)

    r = fn(it, *args)
    if isinstance(it, Sequence):
        return r

    try:
        # Ensure that slice is supported by prefetching 1st item
        first_or_none = *islice(r, 1),
    except TypeError:
        return fallback_fn(it, *args)
    else:
        return chain(first_or_none, r)


# ----------------------------------------------------------------------------


def window_hint(it, size):
    return len(it) + 1 - size


def chunk_hint(it, size):
    return len(range(0, len(it), size))


def _sliced_windowed(s: SupportsSlice[_T], size: int) -> Iterator[_T]:
    indices = range(len(s) + 1)
    slices = map(slice, indices[:-size], indices[size:])
    return map(s.__getitem__, slices)


def _windowed(it: Iterable[_T], size: int) -> Iterator[tuple[_T, ...]]:
    if size == 1:  # Trivial case
        return zip(it)

    it = iter(it)
    w = deque(islice(it, size), maxlen=size)

    if len(w) != size:
        return iter(())
    return map(tuple, chain([w], map(w.__iadd__, zip(it))))


def _sliced(s: SupportsSlice[_T], size: int) -> Iterator[_T]:
    indices = range(len(s) + size)
    slices = map(slice, indices[::size], indices[size::size])
    return map(s.__getitem__, slices)


def _chunked(it: Iterable[_T], size: int) -> Iterator[tuple[_T, ...]]:
    if size == 1:  # Trivial case
        return zip(it)

    # TODO: python 3.12 - `itertools.batched(it, size)`
    fetch_chunk = partial(islice, iter(it), size)
    chunks = iter(fetch_chunk, None)
    return iter(map(tuple, chunks).__next__, ())  # type: ignore[arg-type]


# ---------------------------------------------------------------------------


@overload
def windowed(it: SupportsSlice[_T], size: int) -> Iterator[_T]:
    ...


@overload
def windowed(it: Iterable[_T], size: int) -> Iterator[tuple[_T, ...]]:
    ...


def windowed(it, size):
    """Retrieve overlapped windows from iterable.
    Tries to use slicing if possible.

    >>> [*windowed(range(6), 3)]
    [range(0, 3), range(1, 4), range(2, 5), range(3, 6)]

    >>> [*windowed(iter(range(6)), 3)]
    [(0, 1, 2), (1, 2, 3), (2, 3, 4), (3, 4, 5)]
    """
    return _dispatch(_windowed, _sliced_windowed, it, size)


@overload
def chunked(__it: SupportsSlice[_T], size: int) -> Iterator[_T]:
    ...


@overload
def chunked(__it: Iterable[_T], size: int) -> Iterator[tuple[_T, ...]]:
    ...


def chunked(it, size):
    """
    Splits iterable to chunks of at most size items each.
    Uses slicing if possible.
    Each next() on result will advance passed iterable to size items.

    >>> [*chunked(range(10), 3)]
    [range(0, 3), range(3, 6), range(6, 9), range(9, 10)]

    >>> [*chunked(iter(range(10)), 3)]
    [(0, 1, 2), (3, 4, 5), (6, 7, 8), (9,)]
    """
    return _dispatch(_chunked, _sliced, it, size)


# ----------------------------------------------------------------------------


def _deiter(q: deque[_T]) -> Iterator[_T]:
    # Same as iter_except(q.popleft, IndexError) from docs of itertools
    try:
        while True:
            yield q.popleft()
    except IndexError:
        return


def ichunked(it: Iterable[_T], size: int) -> Iterator[Iterator[_T]]:
    """Split iterable to chunks of at most size items each.

    Does't consume items from passed iterable to return complete chunk
    unlike chunked, as yields iterators, not sequences.

    >>> s = ichunked(range(10), 3)
    >>> len(s)
    4
    >>> [[*chunk] for chunk in s]
    [[0, 1, 2], [3, 4, 5], [6, 7, 8], [9]]
    """
    if size == 1:  # Trivial case
        yield from map(iter, zip(it))  # type: ignore[arg-type]
        return

    it = iter(it)
    while head := deque(islice(it, 1)):
        # Remaining chunk items
        body = islice(it, size - 1)

        # Cache for not-yet-consumed
        tail = deque[_T]()

        # Include early fetched item into chunk
        yield chain(_deiter(head), body, _deiter(tail))

        # Advance and fill internal cache, expand tail with items from body
        tail += body


# ----------------------------------------------------------------------------


def ilen(iterable: Iterable) -> int:
    """Return number of items in *iterable*.

    This consumes iterable, so handle with care.
    See `more_itertools.ilen`.
    """
    counter = count()
    deque(zip(iterable, counter), maxlen=0)
    return next(counter)


def eat(iterable: Iterable, daemon: bool = False) -> None:
    """Consume iterable, daemonize if needed (move to background thread)"""
    if daemon:
        threading.Thread(target=deque, args=(iterable, 0), daemon=True).start()
    else:
        deque(iterable, 0)  # Same as `more_itertools.consume(..., n=None)`


def roundrobin(*iterables: Iterable[_T]) -> Iterator[_T]:
    """roundrobin('ABC', 'D', 'EF') --> A D E B F C"""
    iters = cycle(iter(it) for it in iterables)
    for pending in range(len(iterables) - 1, -1, -1):
        yield from map(next, iters)
        iters = cycle(islice(iters, pending))



from __future__ import annotations

__all__ = [
    'BottleneckResidualBlock',
    'DenseBlock',
    'DenseDelta',
    'ResNeXtBlock',
    'ResidualBlock',
    'SplitAttention',
    'SqueezeExcitation',
    'mbconv',
    'mobilenet_v2_block',
    'mobilenet_v3_block',
    'resnest_block',
]

import torch
from einops.layers.torch import Rearrange, Reduce
from torch import nn
from torchvision.ops.stochastic_depth import StochasticDepth

from .aggregates import Cat, Ensemble, Gate, Residual
from .context import ConvCtx
from .util import ActivationFn, NameMixin, round8

# --------------------------------- densenet ---------------------------------


class DenseBlock(nn.ModuleList):
    expansion = 4
    efficient = True
    __constants__ = ['step', 'depth', 'bottleneck']

    def __init__(self,
                 depth: int = 4,
                 step: int = 16,
                 bottleneck: bool = True,
                 ctx: ConvCtx | None = None):
        ctx = ctx or ConvCtx()
        dim_inner = round8(step * self.expansion)

        layers = []
        for _ in range(depth):
            layer: list[nn.Module] = []
            if bottleneck:
                layer += ctx.norm_act_conv(ctx.conv(dim_inner, 1))
            layer += ctx.norm_act_conv(ctx.conv(step, 3))
            layers.append(
                Cat(*layer) if self.efficient else nn
                .Sequential(Cat(), *layer))

        super().__init__(layers)

        self.step = step
        self.depth = depth
        self.bottleneck = bottleneck

    def __repr__(self) -> str:
        dim_in = next(m for m in self.modules()
                      if isinstance(m, nn.modules.conv._ConvNd)).in_channels

        dim_out = self.step * self.depth
        if not isinstance(self, DenseDelta):
            dim_out = (dim_in + dim_out) * (dim_in != 0)

        line = f'{dim_in}, {dim_out}, step={self.step}, depth={self.depth}'
        if self.bottleneck:
            line += ', bottleneck=True'
        return f'{type(self).__name__}({line})'

    def base_forward(self, x: torch.Tensor) -> list[torch.Tensor]:
        xs = [x]
        for m in self:
            xs.append(m(xs))
        return xs

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        xs = self.base_forward(x)
        return torch.cat(xs, 1)


class DenseDelta(DenseBlock):
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        xs = self.base_forward(x)
        return torch.cat(xs[1:], 1)  # Omit original x


# ---------------------------------- se-net ----------------------------------


class SqueezeExcitation(NameMixin, Gate):
    def __init__(self,
                 dim: int,
                 ratio: float = 0.25,
                 activation: ActivationFn = nn.SiLU,
                 scale_activation: ActivationFn = nn.Sigmoid):
        dim_inner = round8(dim * ratio)
        super().__init__(
            Reduce('b c h w -> b c', 'mean'),
            nn.Linear(dim, dim_inner, bias=False),
            activation(),
            nn.Linear(dim_inner, dim, bias=False),
            scale_activation(),
            Rearrange('b c -> b c 1 1'),
        )
        self.name = f'{dim}, {dim_inner=}'


# ---------------------------------- resnet ----------------------------------


class ResidualBlock(nn.Sequential):
    """BasicBlock from ResNet-18/34"""
    def __init__(self,
                 dim: int,
                 dropout: float = 0.,
                 ctx: ConvCtx | None = None):
        ctx = ctx or ConvCtx()
        # TODO: Support stride
        super().__init__(
            Residual(
                ctx.conv(dim, 3, bias=False),
                ctx.norm(),
                ctx.activation_(),
                ctx.conv(dim, 3, bias=False),
                ctx.norm(),
                StochasticDepth(dropout, 'row'),
            ),
            ctx.activation_(),
        )


class BottleneckResidualBlock(nn.Sequential):
    """BottleneckBlock from ResNet-50/101/152"""

    # https://arxiv.org/abs/1512.03385
    def __init__(self,
                 dim: int,
                 bn_ratio: float = 0.25,
                 se_ratio: float = 0.25,
                 groups: int | None = 1,
                 dropout: float = 0.,
                 ctx: ConvCtx | None = None):
        ctx = ctx or ConvCtx()
        dim_inner = round8(dim * bn_ratio)
        # TODO: Support stride
        super().__init__(
            Residual(
                ctx.conv(dim_inner, 1, bias=False),
                ctx.norm(),
                ctx.activation_(),
                ctx.conv(dim_inner, 3, groups=groups, bias=False),
                ctx.norm(),
                ctx.activation_(),
                ctx.conv(dim, 1, bias=False),
                ctx.norm(),
                *([SqueezeExcitation(dim, se_ratio, ctx.activation)]
                  if se_ratio else []),
                StochasticDepth(dropout, 'row'),
            ),
            ctx.activation_(),
        )


class ResNeXtBlock(BottleneckResidualBlock):
    def __init__(self,
                 dim: int,
                 se_ratio: float = 0.25,
                 dropout: float = 0.,
                 ctx: ConvCtx | None = None):
        ctx = ctx or ConvCtx()
        # TODO: Support stride
        super().__init__(dim, 0.5, se_ratio, 32, dropout, ctx)


def _mbconv_base(dim: int,
                 dim_inner: int,
                 dim_out: int,
                 stride: int = 1,
                 se_ratio: float = 0.25,
                 ctx: ConvCtx | None = None) -> list[nn.Module]:
    ctx = ctx or ConvCtx()  # or nn.HardSwish

    children: list[nn.Module] = []
    if dim != dim_inner:
        children += [
            ctx.conv(dim_inner, 1, bias=False),
            ctx.norm(),
            ctx.activation_(),
        ]
    children += [
        ctx.conv(dim_inner, 3, groups=dim_inner, bias=False) if stride == 1
        else ctx.conv(dim_inner, stride=stride, groups=dim_inner, bias=False)
    ]
    children += [
        ctx.norm(),
        ctx.activation_(),
    ]
    if se_ratio:
        children += [
            SqueezeExcitation(dim_inner, se_ratio, ctx.activation)
            # SqueezeExcitation(dim_inner, se_ratio, ctx.activation,
            #                   nn.Hardsigmoid)
        ]
    children += [
        ctx.conv(dim_out, 1, bias=False),
        ctx.norm(),
    ]
    return children


def mobilenet_v3_block(dim: int,
                       dim_inner: int,
                       dim_out: int | None = None,
                       stride: int = 1,
                       se_ratio: float = 0.,
                       dropout: float = 0.,
                       ctx: ConvCtx | None = None):
    dim_out = dim_out or dim
    children = _mbconv_base(dim, dim_inner, dim_out, stride, se_ratio, ctx)

    if stride != 1 or dim != dim_out:
        return nn.Sequential(*children)
    if dropout:
        children += [StochasticDepth(dropout, 'row')]
    return Residual(*children)


def mobilenet_v2_block(dim: int,
                       dim_out: int | None = None,
                       stride: int = 1,
                       bn_ratio: float = 6.,
                       dropout: float = 0.,
                       ctx: ConvCtx | None = None):
    ctx = ctx or ConvCtx(activation=nn.ReLU6)
    dim_inner = round8(dim * bn_ratio)
    return mobilenet_v3_block(dim, dim_inner, dim_out, stride, 0., dropout,
                              ctx)


# ------------------------------- efficientnet -------------------------------


def mbconv(dim: int,
           stride: int = 1,
           bn_ratio: float = 3.,
           se_ratio: float = 0.25,
           dropout: float = 0.,
           ctx: ConvCtx | None = None) -> nn.Module:
    """
    According to the original article or MaxViT, this one should be:
    ```
    => x + seq(
        norm,  # ! this is redundant in lack of consecutive gelu
        conv1x1, norm, gelu, conv-dw, norm, gelu,
        se,
        conv1x1,
    )
    ```
    for stride == 1, and:
    ```
    => seq(pool, conv1x1) + seq(
        norm,  # ! this is redundant in lack of consecutive gelu
        conv1x1, norm, gelu, conv-dw-pool, norm, gelu,
        se,
        conv1x1,
    )
    ```
    otherwise.
    But this implementation uses post-norm instead, making MBConv the same as
    mobilenet_v3_block.
    """
    dim_inner = round8(dim * bn_ratio)
    return mobilenet_v3_block(dim, dim_inner, dim, stride, se_ratio, dropout,
                              ctx)


# --------------------------------- resnest ----------------------------------


class SplitAttention(NameMixin, nn.Module):
    """
    Split-Attention (aka Splat) block from ResNeSt.
    If radix == 1, equals to SqueezeExitation block from SENet.
    """
    __constants__ = [*NameMixin.__constants__, 'radix']

    def __init__(self,
                 dim: int,
                 groups: int = 1,
                 radix: int = 2,
                 ctx: ConvCtx | None = None):
        assert dim % (groups * radix) == 0
        ctx = ctx or ConvCtx()
        dim_inner = dim * radix // 4

        super().__init__()
        self.radix = radix
        self.to_radix = Rearrange('b (r gc) h w -> b r gc h w', r=radix)
        self.attn = nn.Sequential(
            # Mean by radix and spatial dims
            Reduce('b r gc h w -> b gc 1 1', 'mean'),

            # Core
            ctx.conv(dim_inner, 1, groups=groups, bias=False),
            ctx.norm(),
            ctx.activation_(),
            ctx.conv(dim * radix, 1, groups=groups, bias=False),

            # Normalize
            Rearrange('b (g r c) 1 1 -> b r (g c)', g=groups, r=radix),
            nn.Sigmoid() if radix == 1 else nn.Softmax(1),
        )
        self.name = f'{dim} -> {dim_inner} -> {dim}x{radix}, groups={groups}'

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        rchw = self.to_radix(x)
        rc = self.attn(x * self.radix if self.radix > 1 else x)
        chw = torch.einsum('b r c h w, b r c -> b c h w', rchw, rc)
        return chw.contiguous()


def resnest_block(dim: int,
                  stride: int = 1,
                  radix: int = 1,
                  groups: int = 1,
                  rate: float = 0.25,
                  dropout: float = 0,
                  ctx: ConvCtx | None = None) -> nn.Module:
    ctx = ctx or ConvCtx()
    # dim_inner = round8(dim * rate) * groups  # TODO
    if stride != 1:
        dim_inner = (dim // 4) * groups
        pool = [ctx.avg_pool(stride)]
    else:
        dim_inner = round8(dim * rate)
        pool = []

    core: list[nn.Module] = [
        ctx.conv(dim_inner, 1, bias=False),
        ctx.norm(),
        ctx.activation_(),
        ctx.conv(dim_inner * radix, 3, groups=groups * radix, bias=False),
        ctx.norm(),
        ctx.activation_(),
        SplitAttention(dim_inner, groups, radix),
        *pool,
        ctx.conv(dim, 1, bias=False),
        ctx.norm(),
        StochasticDepth(dropout, 'row'),
    ]

    residual: nn.Module
    if stride != 1:
        downsample = [
            ctx.avg_pool(stride),
            ctx.conv(dim, 1, bias=False),
            ctx.norm(),
        ]
        residual = Ensemble(core, downsample, mode='sum')
    else:
        residual = Residual(*core)

    return nn.Sequential(residual, ctx.activation_())



from __future__ import annotations

__all__ = [
    'BitFlipNoise', 'ChannelMix', 'ChannelShuffle', 'CutOut', 'DegradeJpeg',
    'DegradeQuality', 'Elastic', 'FlipAxis', 'HsvShift', 'LumaJitter',
    'MaskDropout', 'MultiNoise', 'WarpAffine'
]

from dataclasses import InitVar, dataclass, field
from typing import Any

import cv2
import numpy as np
from scipy.stats import ortho_group

from . import functional as F
from .core import DualStageTransform, ImageTransform, MaskTransform

# ---------------------------------- mixins ----------------------------------


class _LutTransform(ImageTransform):
    def get_lut(self,
                rng: np.random.Generator) -> np.ndarray | list[np.ndarray]:
        raise NotImplementedError

    def image(self, image: np.ndarray, rng: np.random.Generator) -> np.ndarray:
        assert image.dtype == np.uint8

        lut = self.get_lut(rng)
        if isinstance(lut, np.ndarray):
            return cv2.LUT(image.ravel(), lut).reshape(image.shape)

        assert len(lut) == image.shape[2]
        planes = map(cv2.LUT, cv2.split(image), lut)
        return cv2.merge([*planes]).reshape(image.shape)


# ---------------------------------- noise ----------------------------------


class AddNoise(ImageTransform):
    """Add uniform[-strength ... +strength] to each item"""
    def __init__(self, strength: float = 0.2) -> None:
        self.strength = int(strength * 255)

    def __repr__(self) -> str:
        return f'{type(self).__name__}(strength={self.strength})'

    def image(self, image: np.ndarray, rng: np.random.Generator) -> np.ndarray:
        assert image.dtype == np.uint8

        res = rng.integers(
            -self.strength, self.strength, size=image.shape, dtype='i2')
        res += image

        return res.clip(0, 255).astype('u1')


class MultiNoise(ImageTransform):
    """Multiply uniform[1 - strength ... 1 + strength] to each item"""
    def __init__(self, strength: float = 0.5) -> None:
        self.low = max(0, 1 - strength)
        self.high = 1 + strength

    def __repr__(self) -> str:
        return f'{type(self).__name__}(low={self.low}, high={self.high})'

    def image(self, image: np.ndarray, rng: np.random.Generator) -> np.ndarray:
        assert image.dtype == np.uint8

        res = rng.random(image.shape, dtype='f4')
        res *= self.high - self.low
        res += self.low
        res *= image  # Multiply

        return res.clip(0, 255).astype('u1')


@dataclass
class BitFlipNoise(ImageTransform):
    bitdepth: int = 4

    def image(self, image: np.ndarray, rng: np.random.Generator) -> np.ndarray:
        assert image.dtype.kind == 'u'
        planes = 8 * image.dtype.itemsize

        if self.bitdepth >= planes:
            return image

        high_flip = 1 << (planes - self.bitdepth)
        bitmask = (1 << planes) - high_flip

        res = image & bitmask
        res += rng.integers(high_flip, size=image.shape, dtype=image.dtype)
        return res


# ----------------------------- color alteration -----------------------------


class ChannelShuffle(ImageTransform):
    def __repr__(self) -> str:
        return f'{type(self).__name__}()'

    def image(self, image: np.ndarray, rng: np.random.Generator) -> np.ndarray:
        assert image.ndim == 3
        return rng.permutation(image, axis=-1)


@dataclass
class ChannelMix(ImageTransform):
    intensity: tuple[float, float] = (0.5, 1.5)

    def image(self, image: np.ndarray, rng: np.random.Generator) -> np.ndarray:
        assert image.ndim == 3
        assert image.dtype == np.uint8
        image = image.astype('f4')

        num_channels = image.shape[-1]
        mat = ortho_group.rvs(num_channels, random_state=rng).astype('f4')

        mat *= rng.uniform(*self.intensity)
        lumat = np.full((num_channels, num_channels), 1 / num_channels)

        image = image @ ((np.eye(num_channels) - lumat) @ mat + lumat)

        return image.clip(0, 255).astype('u1')


@dataclass
class LumaJitter(_LutTransform):
    brightness: tuple[float, float] = (-0.2, 0.2)
    contrast: tuple[float, float] = (0.8, 1.2)

    def get_lut(self, rng: np.random.Generator) -> np.ndarray:
        lut = np.arange(256, dtype='f4')

        lut += 256 * rng.uniform(*self.brightness)
        lut = (lut - 128) * rng.uniform(*self.contrast) + 128

        return lut.clip(0, 255).astype('u1')


@dataclass
class GammaJitter(_LutTransform):
    """Alters gamma from [1/(1+gamma) ... 1+gamma]"""
    gamma: float = 0.2

    def __post_init__(self):
        assert self.gamma >= 0

    def get_lut(self, rng: np.random.Generator) -> np.ndarray:
        lut = np.linspace(0, 1, num=256, dtype='f4')

        max_gamma = 1 + self.gamma
        lut **= rng.uniform(1 / max_gamma, max_gamma)
        lut *= 255

        return lut.clip(0, 255).astype('u1')


@dataclass
class HsvShift(_LutTransform):
    max_shift: int = 20

    def get_lut(self, rng: np.random.Generator) -> list[np.ndarray]:
        hue, sat, val = rng.uniform(-self.max_shift, self.max_shift, size=3)
        ramp = np.arange(256, dtype='i2')
        luts = (
            (ramp + hue) % 180,
            (ramp + sat).clip(0, 255),
            (ramp + val).clip(0, 255),
        )
        return [lut.astype('u1') for lut in luts]

    def image(self, image: np.ndarray, rng: np.random.Generator) -> np.ndarray:
        assert image.ndim == image.shape[-1] == 3

        image = cv2.cvtColor(image, cv2.COLOR_RGB2HSV)
        image = super().image(image, rng)
        return cv2.cvtColor(image, cv2.COLOR_HSV2RGB)


# ------------------------------- compression -------------------------------


@dataclass
class DegradeJpeg(ImageTransform):
    quality: tuple[int, int] = (0, 15)

    def image(self, image: np.ndarray, rng: np.random.Generator) -> np.ndarray:
        quality = int(rng.integers(*self.quality))
        _, buf = cv2.imencode('.jpg', image,
                              (cv2.IMWRITE_JPEG_QUALITY, quality))
        return cv2.imdecode(buf, cv2.IMREAD_UNCHANGED).reshape(image.shape)


@dataclass
class DegradeQuality(ImageTransform):
    scale: tuple[float, float] = (0.25, 0.5)
    modes: tuple[str, ...] = ('NEAREST', 'LINEAR', 'INTER_CUBIC', 'AREA')

    def image(self, image: np.ndarray, rng: np.random.Generator) -> np.ndarray:
        shape = image.shape
        scale = rng.uniform(*self.scale)

        # downscale
        mode = getattr(cv2, f'INTER_{rng.choice(self.modes)}')
        image = cv2.resize(image, None, fx=scale, fy=scale, interpolation=mode)

        # upscale
        mode = getattr(cv2, f'INTER_{rng.choice(self.modes)}')
        image = cv2.resize(image, shape[1::-1], interpolation=mode)
        return image.reshape(shape)


# ----------------------------- mask alteration -----------------------------


@dataclass
class MaskDropout(MaskTransform):
    """
    Drops redundant pixels for each class,
    so that np.bincount(mask.ravel()) <= alpha * mask.size
    """

    alpha: float
    ignore_index: int = -1

    def mask(self, mask: np.ndarray, rng: np.random.Generator) -> np.ndarray:
        return F.mask_dropout(
            mask, rng, alpha=self.alpha, ignore_index=self.ignore_index)


# --------------------------------- geometry ---------------------------------


@dataclass
class FlipAxis(DualStageTransform):
    """
    Flips image/mask vertically/horizontally & rotate by 90 at random.
    In non-isotropic mode (default) flips only horizontally
    """

    isotropic: bool = False

    def prepare(self, rng: np.random.Generator, /, **_) -> dict[str, Any]:
        ud, lr, rot90 = rng.integers(2, size=3)
        if not self.isotropic:
            ud = rot90 = 0
        return {'ud': ud, 'lr': lr, 'rot90': rot90}

    def image(self, image: np.ndarray, **params) -> np.ndarray:
        return F.flip(image, **params)

    def mask(self, mask: np.ndarray, **params) -> np.ndarray:
        return F.flip(mask, **params)


@dataclass
class WarpAffine(DualStageTransform):
    angle: float = 180
    skew: float = 0.5
    scale: tuple[float, float] = (1.0, 1.0)
    inter: InitVar[str] = 'LINEAR'
    _inter: int = field(init=False)

    def __post_init__(self, inter: str):
        self._inter = getattr(cv2, f'INTER_{inter}')

    def prepare(self, rng: np.random.Generator, /, **_) -> dict[str, Any]:
        return {
            'skew': rng.uniform(-self.skew, self.skew),
            'angle': rng.uniform(-self.angle, self.angle),
            'scale': rng.uniform(*self.scale),
        }

    def image(self, image: np.ndarray, **params) -> np.ndarray:
        return F.affine(image, **params, inter=self._inter)

    def mask(self, mask: np.ndarray, **params) -> np.ndarray:
        return F.affine(mask, **params, inter=cv2.INTER_NEAREST)


@dataclass
class Elastic(DualStageTransform):
    """Elastic deformation of image

    Parameters:
    - scale - max shift for each pixel
    - sigma - size of gaussian kernel
    """

    scale: float = 1
    sigma: float = 50
    inter: InitVar[str] = 'LINEAR'
    _inter: int = field(init=False)

    def __post_init__(self, inter: str):
        self._inter = getattr(cv2, f'INTER_{inter}')

    def prepare(self,
                rng: np.random.Generator,
                /,
                image: np.ndarray | None = None,
                **_) -> dict[str, Any] | None:
        if image is None:
            return None
        offsets = rng.random((2, *image.shape[:2]), dtype='f4')
        offsets *= self.scale * 2
        offsets -= self.scale

        for dim, (off, size) in enumerate(zip(offsets, image.shape[:2])):
            shape = np.where(np.arange(2) == dim, size, 1)
            off += np.arange(size).reshape(shape)  # noqa: PLW2901
            cv2.GaussianBlur(off, (17, 17), self.sigma, dst=off)
        return {'offsets': offsets[::-1]}

    def _apply(self, image: np.ndarray, inter: int, **params) -> np.ndarray:
        map_x, map_y = params['offsets']
        return cv2.remap(
            image, map_x, map_y, inter, borderMode=cv2.BORDER_REFLECT_101)

    def image(self, image: np.ndarray, **params) -> np.ndarray:
        return self._apply(image, self._inter, **params)

    def mask(self, mask: np.ndarray, **params) -> np.ndarray:
        return self._apply(mask, cv2.INTER_NEAREST, **params)


@dataclass
class CutOut(ImageTransform):
    max_holes: int = 80
    size: int = 8
    fill_value: int = 0

    def image(self, image: np.ndarray, rng: np.random.Generator) -> np.ndarray:
        num_holes = rng.integers(self.max_holes)
        if not num_holes:
            return image

        anchors = rng.integers(0, image.shape[:2], size=(num_holes, 2))

        # [N, dims, (min, max)]
        holes = anchors[:, :, None] + [-self.size // 2, self.size // 2]
        holes = holes.clip(0, np.array(image.shape[:2])[:, None])

        image = image.copy()
        for (y0, y1), (x0, x1) in holes:
            image[y0:y1, x0:x1] = self.fill_value
        return image



__all__ = [
    'Bias2d', 'BlurPool2d', 'Conv2dWs', 'Decimate2d', 'Noise', 'Upscale2d'
]

from string import ascii_lowercase
from typing import Optional

import torch
import torch.nn.functional as TF
from torch import nn

from .. import functional as F


class Noise(nn.Module):
    __constants__ = ['std']

    def __init__(self, std: float):
        super().__init__()
        self.std = std

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        if not self.training:
            return x
        return torch.empty_like(x).normal_(std=self.std).add_(x)

    def extra_repr(self) -> str:
        return f'std={self.std}'


class Bias2d(nn.Module):
    def __init__(self,
                 dim: int,
                 *size: int,
                 device: Optional[torch.device] = None):
        super().__init__()
        assert len(size) == 2
        self.bias = nn.Parameter(torch.empty(1, dim, *size, device=device))
        self.reset_parameters()

    def reset_parameters(self):
        nn.init.normal_(self.bias)

    def extra_repr(self) -> str:
        _, dim, *space = self.bias.shape
        return f'features={dim}, size={tuple(space)}'

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        bias = self.bias
        size = [x.shape[2], x.shape[3]]

        if torch.jit.is_tracing() or bias.shape[2:] != size:
            # Stretch to input size
            bias = TF.interpolate(
                bias, size, mode='bicubic', align_corners=False)

        return x + bias


class Decimate2d(nn.Module):
    __constants__ = ['stride']

    def __init__(self, stride: int = 2):
        super().__init__()
        self.stride = stride

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return x[:, :, ::self.stride, ::self.stride]

    def extra_repr(self) -> str:
        return f'stride={self.stride}'


class Upscale2d(nn.Module):
    """Upsamples input tensor in `scale` times.
    Use as inverse for `nn.Conv2d(kernel=3, stride=2)`.

    There're 2 different methods:

    - Pixels are thought as squares. Aligns the outer edges of the outermost
      pixels.
      Used in `torch.nn.Upsample(align_corners=True)`.

    - Pixels are thought as points. Aligns centers of the outermost pixels.
      Avoids the need to extrapolate sample values that are outside of any of
      the existing samples.
      In this mode doubling number of pixels doesn't exactly double size of the
      objects in the image.

    This module implements the second way (match centers).
    New image size will be computed as follows:
        `destination size = (source size - 1) * scale + 1`

    For comparison see [here](http://entropymine.com/imageworsener/matching).
    """
    __constants__ = ['stride']

    def __init__(self, stride: int = 2):
        super().__init__()
        self.stride = stride

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return F.upscale2d(x, self.stride)

    def extra_repr(self):
        return f'stride={self.stride}'


class Conv2dWs(nn.Conv2d):
    """
    [Weight standartization](https://arxiv.org/pdf/1903.10520.pdf).
    Better use with GroupNorm(32, features).
    """
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return F.conv2d_ws(x, self.weight, self.bias, self.stride,
                           self.padding, self.dilation, self.groups)


# --------------------------------- blurpool ---------------------------------


def _pascal_triangle(n: int) -> list[int]:
    values = [1]
    for _ in range(n - 1):
        values = [a + b for a, b in zip([*values, 0], [0, *values])]
    return values[:n]


def _outer_mul(*ts: torch.Tensor) -> torch.Tensor:
    assert all(t.ndim == 1 for t in ts)
    letters = ascii_lowercase[:len(ts)]
    return torch.einsum(','.join(letters) + ' -> ' + letters, *ts)


class BlurPool2d(nn.Conv2d):
    def __init__(self,
                 dim: int,
                 kernel: int = 4,
                 stride: int = 2,
                 padding: int = 1,
                 padding_mode: str = 'reflect'):
        super().__init__(dim, dim, kernel, stride, padding, 1, dim, False,
                         padding_mode)
        del self.weight
        self.register_buffer(
            'weight', torch.empty(dim, 1, kernel, kernel), persistent=False)
        self.reset_parameters()

    def reset_parameters(self) -> None:
        if not self.in_channels:
            return

        weights = [
            torch.as_tensor(_pascal_triangle(k)).float()
            for k in self.kernel_size
        ]
        weight = _outer_mul(*weights)
        weight /= weight.sum()

        self.weight.copy_(weight, non_blocking=True)



__all__ = ['Svg']

from collections.abc import Iterator, Sequence
from pathlib import Path

import cv2
import numpy as np
from lxml.builder import ElementMaker
from lxml.etree import ElementTree, indent, tostring

_SVG_NS = 'http://www.w3.org/2000/svg'
_SCRIPT = """\
const path = location.pathname.split("/").pop();
const url = path.substr(0, path.lastIndexOf(".")) + ".jpg";

let svg = document.documentElement;
svg.getElementsByTagName("image")[0].setAttribute("href", url);

let i = new Image();
i.onload = () => {
    svg.setAttribute("height", i.height);
    svg.setAttribute("width", i.width);
};
i.src = url;

for (let group of svg.getElementsByTagName("g")) {
    group.setAttribute("fill", "none");
};
"""


def hsv_colors(count: int) -> Iterator[str]:
    for hue in np.linspace(0, 360, num=count, endpoint=False, dtype='int32'):
        yield f'hsl({hue},100%,50%)'


class Svg:
    """Converts raster mask (2d numpy.ndarray of integers) to SVG-file.

    Parameters:
    - labels - list of all label names (0th is unlabeled),
      i-th name will be assigned to (i+1)th index.

    Usage:
    ```
    mask = cv2.imread('sample.png', cv2.IMREAD_GRAYSCALE)
    Svg(mask, ['pos', 'neg']).save('sample.svg')
    ```
    """
    def __init__(self, mask: np.ndarray, labels: Sequence[str]):
        e = ElementMaker()

        groups = []
        for uniq in np.unique(mask.ravel()):
            if uniq == 0:  # skip background
                continue
            m = (mask == uniq).astype('u1')
            contours = cv2.findContours(m, cv2.RETR_CCOMP,
                                        cv2.CHAIN_APPROX_TC89_L1)[-2]
            polygons = (
                e.polygon(points=' '.join(map(str, contour.ravel())))
                for contour in contours if len(contour) >= 3)
            groups.append(e.g(*polygons, label=labels[uniq - 1]))

        root = e.svg(
            e.image(),
            *groups,
            e.script(href='main.js'),
            e.style('@import url(main.css)'),
            xmlns=_SVG_NS)

        indent(root, space='\t')
        self.body = tostring(root, encoding='utf-8')

        maxlen = max(len(c) for c in labels)
        self.labels = [f'{c:{maxlen}s}' for c in labels]

    def save(self, path: Path) -> None:
        path = Path(path)

        if not (script := path.parent / 'main.js').exists():
            script.write_text(_SCRIPT)

        if not (style := path.parent / 'main.css').exists():
            labels = zip(hsv_colors(len(self.labels)), self.labels)
            style.write_text('\n'.join(
                f'.{name} {{ stroke: {color} }}' for color, name in labels))

        path.with_suffix('.svg').write_text(self.body)

    @staticmethod
    def load(path: Path) -> dict[str, list[np.ndarray]]:
        """
        Yields contours, contour is 2d numpy array of shape [count, (x, y)]
        """
        tree = ElementTree()
        tree.parse(path.with_suffix('.svg').as_posix())

        fields = ((node.attrib['label'], (p.attrib['points'] for p in node))
                  for node in tree.getiterator(f'{{{_SVG_NS}}}g'))
        return {
            name:
            [np.fromstring(ll, 'i4', sep=' ').reshape(-1, 2) for ll in lines]
            for name, lines in fields
        }



from glow.api import env


def test_context():
    assert env == {}

    with env(a=1):
        assert env == {'a': 1}
        with env(a=2, b=3):
            assert env == {'a': 1, 'b': 3}
        assert env == {'a': 1}

    assert env == {}


def test_deco():
    @env(a=2, b=3)
    def fn_1():
        return {**env}

    assert env == {}
    assert fn_1() == {'a': 2, 'b': 3}
    assert env == {}

    @env(a=1)
    def fn_2():
        return fn_1()

    assert fn_2() == {'a': 1, 'b': 3}
    assert env == {}


def test_mix():
    @env(a=2, b=3)
    def fn_1():
        return {**env}

    @env(a=1)
    def fn_2():
        return fn_1()

    with env(a=1):
        assert env == {'a': 1}
        assert fn_1() == {'a': 1, 'b': 3}

    with env(b=4):
        assert fn_2() == {'a': 1, 'b': 4}

    assert env == {}



from __future__ import annotations

__all__ = ['sizeof']

import ctypes
import functools
import sys
from collections import abc
from ctypes import pythonapi
from enum import Enum
from inspect import isgetsetdescriptor, ismemberdescriptor
from numbers import Number
from types import FunctionType, ModuleType

import numpy as np

from ._import_hook import when_imported
from ._repr import si_bin

_ZERO_DEPTH_BASES = (str, bytes, bytearray, range, Number, Enum)
_SINGLETONS = (type, bool, FunctionType, ModuleType)

pythonapi._PyObject_GetDictPtr.argtypes = (ctypes.py_object, )
pythonapi._PyObject_GetDictPtr.restype = ctypes.POINTER(ctypes.py_object)


def unique_only(fn):
    """Protection from self-referencing"""
    def wrapper(obj, seen: set[int]) -> int:
        if (id_ := id(obj)) not in seen:
            seen.add(id_)
            return fn(obj, seen)
        return 0

    return functools.update_wrapper(wrapper, fn)


def true_vars(obj) -> dict[str, object] | None:
    """Get instance's vars(), even it's hidden.

    Inspired by [magic_get_dict](https://stackoverflow.com/a/45315745/9868257)

    For more details see:
    - https://utcc.utoronto.ca/~cks/space/blog/python/HowSlotsWorkI
    - https://utcc.utoronto.ca/~cks/space/blog/python/DictoffsetNotes
    """
    tp = type(obj)

    if sys.implementation.name == 'cpython':
        if (ptr := pythonapi._PyObject_GetDictPtr(obj)) and ptr.contents:
            return ptr.contents.value
        return None

        # if offset := tp.__dictoffset__:
        #     if offset < 0:
        #         offset += tp.__sizeof__(obj)  # type: ignore
        #     addr = id(obj) + offset
        #     ptr = ctypes.cast(addr, ctypes.POINTER(ctypes.py_object))
        #     return ptr.contents.value
        # return None

    if hasattr(obj, '__dict__'):
        for cls in tp.__mro__:
            if (attr := vars(cls).get('__dict__')) is not None:
                return (vars(obj) if (isgetsetdescriptor(attr)
                                      or ismemberdescriptor(attr)) else None)
    return None


@functools.singledispatch
@unique_only
def _sizeof(obj, seen: set[int]) -> int:
    if obj is None:
        return 0

    tp = type(obj)  # Real type, may differ from obj.__class__
    if issubclass(tp, _SINGLETONS):
        return 0

    size = sys.getsizeof(obj)
    if issubclass(tp, _ZERO_DEPTH_BASES):
        return size

    size += _sizeof(true_vars(obj), seen)

    if hasattr(obj, '__wrapped__'):  # Unwrap
        size += _sizeof(obj.__wrapped__, seen)
        return size

    if issubclass(tp, dict):
        size += sum(_sizeof(k, seen) for k in obj)
        size += sum(_sizeof(v, seen) for v in obj.values())
    elif issubclass(tp, abc.Collection):
        size += sum(_sizeof(item, seen) for item in obj)

    if hasattr(obj, '__slots__'):
        size += sum(
            _sizeof(getattr(obj, slot, None), seen)
            for cls in tp.__mro__ if (slots := getattr(cls, '__slots__', ()))
            for slot in ([slots] if isinstance(slots, str) else slots))

    return size


@unique_only
def _sizeof_numpy(obj, seen: set[int]) -> int:
    return _sizeof(obj.base, seen) + sys.getsizeof(obj)


@unique_only
def _sizeof_torch(obj, _: set[int]) -> int:
    size = sys.getsizeof(obj)
    if not obj.is_cuda:
        size += obj.numel() * obj.element_size()
    return size  # TODO: fix if useless when grads are attached


_sizeof.register(np.ndarray, _sizeof_numpy)
when_imported('torch')(
    lambda torch: _sizeof.register(torch.Tensor, _sizeof_torch))


def sizeof(obj) -> int:
    """Computes size of object, no matter how complex it is.

    Inspired by
    [PySize](https://github.com/bosswissam/pysize/blob/master/pysize.py)
    """
    return si_bin(_sizeof(obj, set()))



from __future__ import annotations

__all__ = ['as_actor', 'coroutine']

from collections import Counter, deque
from collections.abc import Callable, Generator, Hashable, Iterable, Iterator
from functools import update_wrapper
from threading import Lock
from typing import TypeVar, cast

import wrapt

from ._more import _deiter

_T = TypeVar('_T')
_R = TypeVar('_R')
_F = TypeVar('_F', bound=Callable[..., Generator])


def coroutine(fn: _F) -> _F:
    def wrapper(*args, **kwargs):
        coro = fn(*args, **kwargs)
        coro.send(None)
        return coro

    return cast(_F, update_wrapper(wrapper, fn))


class _Sync(wrapt.ObjectProxy):
    def __init__(self, wrapped):
        super().__init__(wrapped)
        self._self_lock = Lock()

    def _call(self, op, *args, **kwargs):
        with self._self_lock:
            return op(*args, **kwargs)

    def __next__(self):
        return self._call(self.__wrapped__.__next__)

    def send(self, item):
        return self._call(self.__wrapped__.send, item)

    def throw(self, typ, val=None, tb=None):
        return self._call(self.__wrapped__.throw, typ, val, tb)

    def close(self):
        return self._call(self.__wrapped__.close)


def threadsafe_iter(fn: _F) -> _F:
    def wrapper(*args, **kwargs):
        gen = fn(*args, **kwargs)
        return _Sync(gen)

    return cast(_F, update_wrapper(wrapper, fn))


@threadsafe_iter
@coroutine
def summary() -> Generator[None, Hashable | None, None]:
    # ? delete this or find use case
    state = Counter[Hashable]()
    while True:
        key = yield
        if key is None:
            state.clear()
        else:
            state[key] += 1
            print(dict(state), flush=True, end='\r')


@threadsafe_iter
@coroutine
def as_actor(
    transform: Callable[[Iterable[_T]], Iterator[_R]],
) -> Generator[_R, _T, None]:
    buf = deque[_T]()
    gen = transform(_deiter(buf))  # infinite

    # shortcuts
    buf_append, gen_next = buf.append, gen.__next__

    x = yield  # type: ignore[misc]  # preseed coroutine
    while True:
        buf_append(x)
        x = yield gen_next()



__all__ = ['plot_model']

import functools
from collections.abc import Iterable, Iterator, Mapping
from contextlib import ExitStack

import graphviz
import torch
from torch import nn
from torch.autograd import Function

from .. import mangle, si

# TODO: Still buggy, continue research/refactor


def id_(x) -> str:
    if hasattr(x, 'variable'):
        x = x.variable
    addr = x.storage().data_ptr() if isinstance(x, torch.Tensor) else id(x)
    return hex(addr)


def flatten(xs) -> Iterator[torch.Tensor]:
    if xs is None:
        return
    if isinstance(xs, torch.Tensor):
        yield xs
        return
    if isinstance(xs, Iterable):
        if isinstance(xs, Mapping):
            xs = xs.items()
        for x in xs:
            yield from flatten(x)
        return
    raise TypeError(f'Unsupported argument type: {type(xs)}')


def sized(var: torch.Tensor):
    if max(var.shape) == var.numel():
        return f'{tuple(var.shape)}'
    return f'{tuple(var.shape)}\n{si(var.numel())}'


class Builder:
    def __init__(self,
                 inputs: set[str],
                 params: dict[str, str],
                 *,
                 nesting: bool = True,
                 variables: bool = True):
        self.inputs = inputs
        self.params = params
        self.nesting = nesting
        self.variables = variables

        self._mangle = mangle()
        self._memo: dict[str, str] = {}
        self._shapes: dict[Function, str] = {}
        root = graphviz.Digraph(
            name='root',
            graph_attr={
                'rankdir': 'LR',
                'newrank': 'true',
                'color': 'lightgrey',
            },
            edge_attr={
                'labelfloat': 'true',
            },
            node_attr={
                'shape': 'box',
                'style': 'filled',
                'fillcolor': 'lightgrey',
                'fontsize': '12',
                'height': '0.2',
                'ranksep': '0.1',
            },
        )
        self.stack = [root]

    def _add_op_node(self, grad_id: str, grad: Function):
        label = type(grad).__name__.replace('Backward', '')
        if grad in self._shapes:
            label = f'{label}\n=> {tuple(self._shapes[grad])}'
        self.stack[-1].node(grad_id, label)

    def _add_var_node(self, var_id: str, var: torch.Tensor):
        label_ = []
        if param_name := self.params.get(var_id):
            root = self.stack[-1]
            parts = param_name.split('.')
            label_.append(parts[-1] if self.nesting else '.'.join(parts[1:]))
        else:
            root = self.stack[0]  # unnamed, that's why external

        label = '\n'.join([*label_, var_id, sized(var)])
        color = 'yellow' if var_id in self.inputs else 'lightblue'
        root.node(var_id, label, fillcolor=color)

    def _traverse_saved(self, grad_id: str, *tensors):
        tensors = *(v for v in tensors if isinstance(v, torch.Tensor)),
        if not tensors:
            return
        s_ctx = self.stack[-1].subgraph()
        assert s_ctx is not None
        with s_ctx as s:
            s.attr(rank='same')
            for var in tensors:
                var_id = id_(var)
                if var_id not in self._memo:
                    label = f'{var_id}\n{sized(var)}'
                    s.node(var_id, label, fillcolor='orange')
                s.edge(var_id, grad_id)

    def _traverse(self, grad: Function, depth: int = 0):
        if grad is None or (grad_id := id_(grad)) in self._memo:
            return

        root = self.stack[-1]
        self._memo[grad_id] = head = root.name
        if hasattr(grad, 'variable'):
            # Has variable, so it's either Parameter or Variable
            self._add_var_node(grad_id, grad.variable)
            yield (depth - 1, None, grad)
            return

        # Doesn't have variable, so it's "operation"
        self._add_op_node(grad_id, grad)

        # TODO : add merging of tensors with same data
        if self.variables and hasattr(grad, 'saved_tensors'):
            self._traverse_saved(grad_id, *(grad.saved_tensors or ()))

        for grad_next, _ in getattr(grad, 'next_functions', ()):
            if grad_next is None:
                continue
            yield from self._traverse(grad_next, depth + 1)

            next_id = id_(grad_next)
            tail = self._memo.get(next_id)
            if tail is not None and head is not None and not (
                    head.startswith(tail) or tail.startswith(head)):
                yield (depth, grad_next, grad)  # leafs, yield for depth-check
                continue

            name = self.params.get(next_id)
            if self.nesting and name and name.rpartition('.')[0] == head:
                s_ctx = root.subgraph()
                assert s_ctx is not None
                with s_ctx as s:
                    s.attr(rank='same')
                    s.edge(next_id, grad_id)  # same module, same rank
            else:
                self.stack[0].edge(next_id, grad_id)

    def _mark(self, ts):
        edges = []
        for t in flatten(ts):
            if t.grad_fn is not None:
                self._shapes[t.grad_fn] = t.shape  # type: ignore[assignment]
                edges += self._traverse(t.grad_fn)
        if not edges:
            return

        max_depth = max(depth for depth, *_ in edges) + 1
        for depth, tail, head in edges:  # inter-module edges
            if tail is not None:
                minlen = f'{max_depth - depth}' if self.nesting else None
                self.stack[0].edge(id_(tail), id_(head), minlen=minlen)

    def forward_pre(self, name, module, xs):
        self._mark(xs)
        # -------- start node --------
        if not self.nesting:
            return
        scope = graphviz.Digraph(name=self._mangle(name))
        scope.attr(label=f'{name.split(".")[-1]}:{type(module).__name__}')
        self.stack.append(scope)

    def forward(self, module, _, ys):
        self._mark(ys)
        if not self.nesting:
            return
        cluster = self.stack.pop()
        cluster.name = f'cluster_{cluster.name}'
        self.stack[-1].subgraph(cluster)
        # -------- end node --------


def plot_model(model: nn.Module,
               *input_shapes: tuple[int, ...],
               device='cpu',
               nesting: bool = True,
               variables: bool = False):
    """Produces Graphviz representation of PyTorch autograd graph

    Blue nodes are the Variables that require grad, orange are Tensors
    saved for backward in torch.autograd.Function
    """
    inputs = [
        torch.zeros(1, *s, device=device, requires_grad=True)
        for s in input_shapes
    ]
    params = model.state_dict(prefix='root.', keep_vars=True)
    hk = Builder(
        {id_(var) for var in inputs},
        {id_(var): name for name, var in params.items()},
        nesting=nesting,
        variables=variables,
    )
    with ExitStack() as stack:
        for name, m in model.named_modules(prefix='root'):
            stack.callback(
                m.register_forward_pre_hook(
                    functools.partial(hk.forward_pre, name)).remove)
            stack.callback(m.register_forward_hook(hk.forward).remove)
        model(*inputs)

    dot = hk.stack.pop()
    assert not hk.stack

    dot.filename = getattr(model, 'name', type(model).__qualname__)
    dot.directory = 'graphs'
    dot.format = 'svg'

    size_min = 12
    scale_factor = .15
    size = max(size_min, len(dot.body) * scale_factor)

    dot.graph_attr.update(size=f'{size},{size}')
    dot.render(cleanup=True)
    return dot



from collections.abc import Callable, Iterable, Iterator, Mapping
from concurrent.futures import Executor
from typing import Generic, Protocol, TypeVar, overload

_T = TypeVar('_T')
_T1_ctr = TypeVar('_T1_ctr', contravariant=True)
_T2_ctr = TypeVar('_T2_ctr', contravariant=True)
_T3_ctr = TypeVar('_T3_ctr', contravariant=True)
_K_co = TypeVar('_K_co', covariant=True)
_R_co = TypeVar('_R_co', covariant=True)


class _Callable1(Generic[_T1_ctr, _R_co], Protocol):
    def __call__(self, __1: _T1_ctr, /) -> _R_co:
        ...


class _Callable2(Generic[_T1_ctr, _T2_ctr, _R_co], Protocol):
    def __call__(self, __1: _T1_ctr, __2: _T2_ctr, /) -> _R_co:
        ...


class _Callable3(Generic[_T1_ctr, _T2_ctr, _T3_ctr, _R_co], Protocol):
    def __call__(self, __1: _T1_ctr, __2: _T2_ctr, __3: _T3_ctr, /) -> _R_co:
        ...


class _Callable4(Generic[_R_co], Protocol):
    def __call__(self, __1, __2, __3, __4, *args) -> _R_co:
        ...


def max_cpu_count(upper_bound: int = ..., mp: bool = ...) -> int:
    ...


def buffered(__iter: Iterable[_T],
             /,
             *,
             latency: int = ...,
             mp: bool | Executor = ...) -> Iterator[_T]:
    ...


def starmap_n(__func: Callable[..., _R_co],
              __iter: Iterable[Iterable],
              /,
              *,
              max_workers: int | None = ...,
              prefetch: int | None = ...,
              mp: bool = ...,
              chunksize: int | None = ...,
              order: bool = ...) -> Iterator[_R_co]:
    ...


@overload
def map_n(__func: _Callable1[_T1_ctr, _R_co],
          __iter1: Iterable[_T1_ctr],
          /,
          *,
          max_workers: int | None = ...,
          prefetch: int | None = ...,
          mp: bool = ...,
          chunksize: int | None = ...,
          order: bool = ...) -> Iterator[_R_co]:
    ...


@overload
def map_n(__f: _Callable2[_T1_ctr, _T2_ctr, _R_co],
          __iter1: Iterable[_T1_ctr],
          __iter2: Iterable[_T2_ctr],
          /,
          *,
          max_workers: int | None = ...,
          prefetch: int | None = ...,
          mp: bool = ...,
          chunksize: int | None = ...,
          order: bool = ...) -> Iterator[_R_co]:
    ...


@overload
def map_n(__f: _Callable3[_T1_ctr, _T2_ctr, _T3_ctr, _R_co],
          __iter1: Iterable[_T1_ctr],
          __iter2: Iterable[_T2_ctr],
          __iter3: Iterable[_T3_ctr],
          /,
          *,
          max_workers: int | None = ...,
          prefetch: int | None = ...,
          mp: bool = ...,
          chunksize: int | None = ...,
          order: bool = ...) -> Iterator[_R_co]:
    ...


@overload
def map_n(__func: _Callable4[_R_co],
          /,
          __iter1: Iterable,
          __iter2: Iterable,
          __iter3: Iterable,
          __iter4: Iterable,
          *__iters: Iterable,
          max_workers: int | None = ...,
          prefetch: int | None = ...,
          mp: bool = ...,
          chunksize: int | None = ...,
          order: bool = ...) -> Iterator[_R_co]:
    ...


def map_n_dict(func: Callable[[_T1_ctr], _R_co],
               obj: Mapping[_K_co, _T1_ctr],
               /,
               *,
               max_workers: int | None = None,
               prefetch: int | None = 2,
               mp: bool = False,
               chunksize: int | None = None) -> dict[_K_co, _R_co]:
    ...



from weakref import WeakValueDictionary

import numpy as np

from glow import chunked, ichunked, windowed


def test_windowed():
    it = windowed(range(5), 3)
    assert len(it) == 3
    assert [*it] == [range(0, 3), range(1, 4), range(2, 5)]
    assert len(it) == 0

    it = windowed(iter(range(5)), 3)
    assert [*it] == [(0, 1, 2), (1, 2, 3), (2, 3, 4)]


def test_chunked():
    it = chunked(range(5), 3)
    assert len(it) == 2
    assert [*it] == [range(0, 3), range(3, 5)]
    assert len(it) == 0

    it = chunked(iter(range(5)), 3)
    assert [*it] == [(0, 1, 2), (3, 4)]


def test_ichunked():
    it = ichunked(range(5), 3)
    assert len([*it]) == 2

    it = ichunked(range(5), 3)
    assert [tuple(c) for c in it] == [(0, 1, 2), (3, 4)]

    it = ichunked(range(5), 3)
    assert [tuple(c) for c in [*it]] == [(0, 1, 2), (3, 4)]


def generator(n, refs):
    heap = {}
    for i, x in enumerate(np.random.rand(n, 4)):
        refs[i] = heap[i] = x
        del x
        # Careful not to keep reference to x
        yield heap.popitem()


def test_windowed_refs():
    refs = WeakValueDictionary()
    assert not refs

    it = generator(10, refs)
    for c in windowed(it, 3):
        assert len(refs) <= 3
        del c
    assert not refs


def test_chunked_refs():
    refs = WeakValueDictionary()
    assert not refs

    it = generator(10, refs)
    for c in chunked(it, 3):
        assert len(refs) <= 3
        del c
    assert not refs


def test_ichunked_refs():
    refs = WeakValueDictionary()
    assert not refs

    it = generator(10, refs)
    for c in ichunked(it, 3):
        for x in c:
            assert len(refs) <= 1
            del x
    assert not refs

    it = generator(10, refs)
    chunks = [*ichunked(it, 3)]
    assert len(refs) == 10
    for i, x in zip(range(10, 0, -1), (x for c in chunks for x in c)):
        assert len(refs) <= i
        del x
    assert not refs



__all__ = ['get_gpu_capability', 'get_gpu_memory_info']

import os
from contextlib import contextmanager
from typing import NamedTuple

import pynvml

from .. import si_bin


@contextmanager
def _nvml():
    pynvml.nvmlInit()
    try:
        yield
    finally:
        pynvml.nvmlShutdown()


def _get_device_handles() -> list:
    devices = os.environ.get('CUDA_VISIBLE_DEVICES')
    if devices is not None:
        indices = [int(dev) for dev in devices.split(',')]
    else:
        *indices, = range(int(pynvml.nvmlDeviceGetCount()))

    return [pynvml.nvmlDeviceGetHandleByIndex(i) for i in indices]


class _GpuState(NamedTuple):
    free: list[int]
    total: list[int]


def get_gpu_capability() -> list[tuple[int, int]]:
    """Gives CUDA capability for each GPU"""
    with _nvml():
        handles = _get_device_handles()
        return [pynvml.nvmlDeviceGetCudaComputeCapability(h) for h in handles]


def get_gpu_memory_info() -> _GpuState:
    """Gives size of free and total VRAM memory for each GPU"""
    with _nvml():
        handles = _get_device_handles()
        *infos, = map(pynvml.nvmlDeviceGetMemoryInfo, handles)

    return _GpuState([si_bin(i.free) for i in infos],
                     [si_bin(i.total) for i in infos])
